import csv
import os
import time
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple

import click
import yaml
from PyInquirer import prompt

from conformity_migration.cloud_accounts import get_cloud_account_adder
from conformity_migration.conformity_api import (
    CloudOneConformityAPI,
    ConformityAPI,
    ConformityError,
    LegacyConformityAPI,
)
from conformity_migration.models import (
    Account,
    AccountDetails,
    Check,
    CommunicationSettings,
    Group,
    Note,
    Profile,
    ReportConfig,
    Rule,
    User,
)

from . import __version__ as tool_version
from .di import (
    app_config,
    c1_conformity_api,
    legacy_conformity_api,
    logger,
    user_config_path,
)
from .utils import str2bool

log = logger()


AccountEnv = Tuple[str, str]


def create_user_config(user_conf_path: Path):
    if user_conf_path.exists():
        recreate_ok = ask_confirmation(
            msg="Configuration file already exists. Do you want to recreate it?",
            ask_if_sure=True,
        )
        if not recreate_ok:
            return

    region_map = app_config()["REGION_API_URL"]
    region_choices = [*region_map.keys(), "Other"]
    legacy_region = ask_choices(
        msg="Legacy Conformity Region", choices=region_choices, default=1
    )
    legacy_api_url = region_map.get(legacy_region)
    if not legacy_api_url:
        legacy_api_url = ask_input(
            "Please input the Legacy Conformity API URL (e.g. https://us-west-2-api.cloudconformity.com/v1:"
        )
    legacy_api_key = ask_input("Legacy Conformity API KEY:", mask_input=True)

    c1_region = ask_choices(
        msg="CloudOne Conformity Region", choices=region_choices, default=1
    )
    c1_api_url = region_map.get(c1_region)
    if not c1_api_url:
        c1_api_url = ask_input(
            "Please input the CloudOne Conformity API URL (e.g. https://us-west-2-api.cloudconformity.com/v1:"
        )
    c1_api_key = ask_input("CloudOne Conformity API KEY:", mask_input=True)

    conf = {
        "CLOUD_ONE_CONFORMITY": {
            "API_KEY": c1_api_key,
            "API_BASE_URL": c1_api_url,
        },
        "LEGACY_CONFORMITY": {
            "API_KEY": legacy_api_key,
            "API_BASE_URL": legacy_api_url,
        },
    }
    with open(user_conf_path, mode="w") as fh:
        return yaml.dump(conf, fh)


def cloud_type_accts_map(accts: List[Account]) -> Dict[str, List[Account]]:
    accts_map: Dict[str, List[Account]] = dict()

    for acct in accts:
        cloud_type_accts = accts_map.setdefault(acct.cloud_type, [])
        cloud_type_accts.append(acct)

    return accts_map


def acct_env_suffix(acct_environment: str) -> str:
    return f" ({acct_environment})" if acct_environment else ""


def empty_legacy_conformity(legacy_api: LegacyConformityAPI):
    empty_conformity(api=legacy_api, conformity_type="Legacy")


def empty_c1_conformity(c1_api: CloudOneConformityAPI):
    empty_conformity(api=c1_api, conformity_type="Cloud One")


def empty_conformity(api: ConformityAPI, conformity_type: str):
    continue_ok = ask_confirmation_with_text_verification(
        msg=f"!!! WARNING !!! This will delete all your {conformity_type} Conformity accounts and configurations. Do you want to continue?",
        verify_text=f"empty {conformity_type} Conformity",
    )
    if not continue_ok:
        return

    log.info("Resetting Organisational Profile")
    api.reset_organisation_profile()

    log.info("Deleting all Custom Profiles")
    for prof in api.get_custom_profiles():
        log.info(f"  -- {prof.name}")
        api.delete_profile(profile_id=prof.profile_id)

    log.info("Deleting all Organisational Report Configs")
    for rconf in api.list_organisation_report_configs():
        log.info(f"  -- {rconf.title}")
        api.delete_report_config(report_conf_id=rconf.report_config_id)

    log.info("Deleting all Communication Settings (Organisation-level)")
    cs_ids = {cs.com_setting_id for cs in api.get_communication_settings(acct_id="")}
    for cs_id in cs_ids:
        log.info(f" -- {cs_id}")
        api.delete_communication_settings(com_setting_id=cs_id)

    groups = api.list_groups()
    for group in groups:
        log.info(f"Deleting Report Configs for group {group.name}")
        for rconf in api.list_group_report_configs(group_id=group.group_id):
            log.info(f"  -- {rconf.title}")
            api.delete_report_config(report_conf_id=rconf.report_config_id)

    log.info("Deleting all Accounts")
    for acct in api.list_accounts():
        env_suffix = acct_env_suffix(acct.environment)
        log.info(f"  -- {acct.name}{env_suffix}")
        api.delete_account(acct_id=acct.account_id)

    log.info("Deleting all Groups")
    for group in groups:
        log.info(f"  -- {group.name}")
        api.delete_group(group_id=group.group_id)


def prompt_initialize_organisation_profile():
    log.info(
        """The Organisation Profile of Cloud One Conformity must be manually initialized first before
this tool can migrate the Organisation Profile settings. Follow these steps to initialize it:
    1. Login to your Cloud One account.
    2. Click Conformity.
    3. On the top navigation bar, click on Profiles.
    4. Under the Default profiles, click on Organisation Profile
    5. Wait for the Rule Settings to load.

    When you are done, please choose Yes below.
"""
    )

    done = False
    while not done:
        done = ask_confirmation(msg="Are you done?", ask_if_sure=True)


def run_migration(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    include_accts=Optional[Set[AccountEnv]],
    exclude_accts=Optional[Set[AccountEnv]],
):
    # this is part of workaround fix for Conformity Public API
    # will initialize Organisation Profile
    prompt_initialize_organisation_profile()

    exec_migration_func(lambda: update_organisation_profile(legacy_api, c1_api))

    exec_migration_func(lambda: add_managed_groups(legacy_api, c1_api))

    log.info("Adding all cloud accounts", flush=True)
    cloud_accts_to_migrate = exec_migration_func(
        lambda: add_cloud_accounts(
            legacy_api=legacy_api,
            c1_api=c1_api,
            include_accts=include_accts,
            exclude_accts=exclude_accts,
        )
    )
    if cloud_accts_to_migrate is None:
        cloud_accts_to_migrate = dict()

    log.info("Retrieving Legacy Conformity Users", flush=True)
    legacy_users = legacy_api.get_all_users()

    log.info("Retrieving CloudOne Conformity Users", flush=True)
    c1_users = c1_api.get_all_users()

    users_to_invite = set(legacy_users).difference(set(c1_users))
    if users_to_invite:
        invite_users(users_to_invite)

    users_to_verify_mobile = [user for user in legacy_users if user.is_mobile_verified]
    if users_to_verify_mobile:
        verify_users_mobile_numbers(users_to_verify_mobile)

    if any([users_to_invite, users_to_verify_mobile]):
        log.info("Retrieving updated list of CloudOne Conformity Users", flush=True)
        c1_users = c1_api.get_all_users()

    exec_migration_func(lambda: create_user_defined_groups(legacy_api, c1_api))

    exec_migration_func(lambda: copy_custom_profiles(legacy_api, c1_api))

    exec_migration_func(lambda: copy_organisation_report_configs(legacy_api, c1_api))

    exec_migration_func(lambda: migrate_all_groups_configs(legacy_api, c1_api))

    c1_org_id = c1_api.get_organisation_id()

    log.info("Copying communication channel settings (organisation-level)", flush=True)
    exec_migration_func(
        lambda: copy_communication_channel_settings(
            legacy_api=legacy_api,
            c1_api=c1_api,
            legacy_acct_id="",
            c1_acct_id="",
            legacy_users=legacy_users,
            c1_users=c1_users,
            c1_org_id=c1_org_id,
        )
    )

    for _, acct_id_map in cloud_accts_to_migrate.items():
        for legacy_acct_id, c1_acct_id in acct_id_map.items():
            exec_migration_func(
                lambda: migrate_account_configurations(
                    legacy_api=legacy_api,
                    c1_api=c1_api,
                    legacy_acct_id=legacy_acct_id,
                    c1_acct_id=c1_acct_id,
                    legacy_users=legacy_users,
                    c1_users=c1_users,
                    c1_org_id=c1_org_id,
                )
            )
            log.info("")


def migrate_all_groups_configs(
    legacy_api: LegacyConformityAPI, c1_api: CloudOneConformityAPI
):
    c1_group_id_map: Dict[Group, str] = {g: g.group_id for g in c1_api.list_groups()}

    for legacy_group in legacy_api.list_groups():
        log.info(
            f"Migrating group configurations for Group={legacy_group.name}, Tags={legacy_group.tags}",
            flush=True,
        )
        legacy_group_id = legacy_group.group_id
        c1_group_id = c1_group_id_map.get(legacy_group)
        if not c1_group_id:
            log.warn(
                f"Can't find corresponding CloudOne group for: Group={legacy_group.name}, Tags={legacy_group.tags}. Cannot migrate it's configurations."
            )
            continue
        exec_migration_func(
            lambda: copy_group_report_configs(
                legacy_api=legacy_api,
                c1_api=c1_api,
                legacy_group_id=legacy_group_id,
                c1_group_id=c1_group_id,  # type: ignore
            )
        )


def update_organisation_profile(
    legacy_api: LegacyConformityAPI, c1_api: CloudOneConformityAPI
):
    log.info("Copying Organisation Profile", flush=True)
    legacy_org_profile = legacy_api.get_organisation_profile(include_rule_settings=True)
    c1_org_profile = c1_api.get_organisation_profile(include_rule_settings=True)

    if c1_org_profile.included_rules:
        overwrite = ask_confirmation_or_auto_overwrite(
            "CloudOne Organisation Profile has configured rules in it. Do you want to overwrite it?",
            ask_if_sure=True,
        )
        if not overwrite:
            return
    c1_api.update_organisation_profile(profile=legacy_org_profile)


def include_exclude_accts(
    legacy_accts: List[Account],
    include_accts=Optional[Set[AccountEnv]],
    exclude_accts=Optional[Set[AccountEnv]],
) -> List[Account]:
    # a value of None means it will not choose any account to include -- all accounts will be added to Cloud One unless in the excluded list
    # a value of an empty set means it will include nothing -- no account will be added to Cloud One
    if include_accts is not None:
        legacy_accts = [
            acct
            for acct in legacy_accts
            if (acct.name, acct.environment) in include_accts
        ]

    # a value of None or of an empty set means the same thing: it will exclude nothing
    if exclude_accts:
        legacy_accts = [
            acct
            for acct in legacy_accts
            if (acct.name, acct.environment) not in exclude_accts
        ]

    return legacy_accts


def add_cloud_accounts(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    include_accts=Optional[Set[AccountEnv]],
    exclude_accts=Optional[Set[AccountEnv]],
) -> Dict[str, Dict[str, str]]:

    c1_accts = c1_api.list_accounts()
    legacy_accts = legacy_api.list_accounts()

    legacy_accts = include_exclude_accts(
        legacy_accts=legacy_accts,
        include_accts=include_accts,
        exclude_accts=exclude_accts,
    )

    legacy_cloud_type_accts_map = cloud_type_accts_map(legacy_accts)
    c1_cloud_type_accts_map = cloud_type_accts_map(c1_accts)

    cloud_accts_to_migrate: Dict[str, Dict[str, str]] = dict()

    for cloud_type, legacy_accts in legacy_cloud_type_accts_map.items():
        acct_adder = get_cloud_account_adder(
            cloud_type=cloud_type, legacy_api=legacy_api, c1_api=c1_api
        )
        if acct_adder is None:
            log.info(f"Does not support {cloud_type.upper()} yet! Skipping it.")
            continue

        log.info(f"Adding {cloud_type.upper()} accounts to CloudOne Conformity:")
        c1_cloud_type_accounts = c1_cloud_type_accts_map.get(cloud_type, [])

        accts_to_migrate: Dict[str, str] = dict()
        cloud_accts_to_migrate[cloud_type] = accts_to_migrate
        for acct in legacy_accts:
            c1_acct_id: str
            exists, c1_acct_id = acct_adder.account_exists(
                c1_accts=c1_cloud_type_accounts, acct=acct
            )
            env_suffix = acct_env_suffix(acct.environment)
            if exists:
                log.info(
                    f"Account {acct.name}{env_suffix} already exists in CloudOne Conformity!"
                )
                if not (
                    ask_confirmation_or_auto_overwrite(
                        "Do you want to migrate configurations for this account (will overwrite existing ones)?",
                        ask_if_sure=True,
                    )
                ):
                    continue
            else:
                log.info(f" --> Account: {acct.name}{env_suffix}")
                c1_acct_id = exec_migration_func(
                    lambda: acct_adder.account_add(acct=acct)  # type: ignore
                )
                if not c1_acct_id:
                    continue

            accts_to_migrate[acct.account_id] = c1_acct_id

    return cloud_accts_to_migrate


def create_custom_profile(
    profile: Profile, legacy_api: LegacyConformityAPI, c1_api: CloudOneConformityAPI
):
    log.info(f"  --> Profile: {profile.name}", flush=True)
    profile_with_rules = legacy_api.get_profile(
        profile_id=profile.profile_id, include_rule_settings=True
    )
    c1_api.create_new_profile(profile=profile_with_rules)


def copy_custom_profiles(
    legacy_api: LegacyConformityAPI, c1_api: CloudOneConformityAPI
):
    log.info("Copying Custom Profiles", flush=True)
    legacy_profiles = legacy_api.get_custom_profiles()
    c1_profiles = c1_api.get_custom_profiles()

    legacy_profiles_set = set(legacy_profiles)
    c1_profiles_to_replace = [p for p in c1_profiles if p in legacy_profiles_set]
    if c1_profiles_to_replace:
        log.info(
            "Found following custom profiles that will be replaced during migration:"
        )
        for c1_profile in c1_profiles_to_replace:
            log.info(f"  - Profile: {c1_profile.name}")
        cont = ask_confirmation_or_auto_overwrite(
            "Continue migrating custom profiles?", ask_if_sure=True
        )
        if not cont:
            return

    for profile in c1_profiles_to_replace:
        # log.info(f" --> Deleting CloudOne profile: {profile.name}")
        c1_api.delete_profile(profile_id=profile.profile_id)

    for profile in legacy_profiles:
        exec_migration_func(
            lambda: create_custom_profile(
                profile=profile, legacy_api=legacy_api, c1_api=c1_api
            )
        )


def check_existing_c1_report_configs(
    c1_api: CloudOneConformityAPI,
    legacy_report_configs: List[ReportConfig],
    c1_report_configs: List[ReportConfig],
    rconf_type: str,
) -> bool:
    rconf_type = rconf_type.capitalize()

    legacy_rconf_set = set(legacy_report_configs)
    c1_rconf_to_replace = [r for r in c1_report_configs if r in legacy_rconf_set]
    if c1_rconf_to_replace:
        log.info(
            f"Found following {rconf_type} Report Configs that will be replaced during migration:"
        )
        for c1_rconf in c1_rconf_to_replace:
            log.info(f"  - Report Config: {c1_rconf.title}")
        cont = ask_confirmation_or_auto_overwrite(
            f"Continue migrating {rconf_type} Report Configs?", ask_if_sure=True
        )
        if not cont:
            return False

    for rconf in c1_rconf_to_replace:
        # log.info(f" --> Deleting CloudOne {rconf_type} Report Config: {rconf.title}")
        c1_api.delete_report_config(report_conf_id=rconf.report_config_id)

    return True


def copy_account_report_configs(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_acct_id: str,
    c1_acct_id: str,
):
    rconf_type = "Account"
    log.info(f"  --> Copying {rconf_type} Report Configs", flush=True)
    legacy_report_configs = legacy_api.list_account_report_configs(
        acct_id=legacy_acct_id
    )
    c1_report_configs = c1_api.list_account_report_configs(acct_id=c1_acct_id)

    cont_migration = check_existing_c1_report_configs(
        c1_api=c1_api,
        legacy_report_configs=legacy_report_configs,
        c1_report_configs=c1_report_configs,
        rconf_type=rconf_type,
    )
    if not cont_migration:
        return

    for report_config in legacy_report_configs:
        log.info(f"    --> Report Config: {report_config.title}")
        exec_migration_func(
            lambda: c1_api.create_account_report_config(
                report_conf=report_config.configuration, acct_id=c1_acct_id
            )
        )


def copy_group_report_configs(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_group_id: str,
    c1_group_id: str,
):
    rconf_type = "Group"
    log.info(f" --> Copying {rconf_type} Report Configs", flush=True)
    legacy_report_configs = legacy_api.list_group_report_configs(
        group_id=legacy_group_id
    )
    c1_report_configs = c1_api.list_group_report_configs(group_id=c1_group_id)

    cont_migration = check_existing_c1_report_configs(
        c1_api=c1_api,
        legacy_report_configs=legacy_report_configs,
        c1_report_configs=c1_report_configs,
        rconf_type=rconf_type,
    )
    if not cont_migration:
        return

    for report_config in legacy_report_configs:
        log.info(f"    --> Report Config: {report_config.title}")
        exec_migration_func(
            lambda: c1_api.create_group_report_config(
                report_conf=report_config.configuration, group_id=c1_group_id
            )
        )


def copy_organisation_report_configs(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
):
    rconf_type = "Organisation"
    log.info(f"Copying {rconf_type} Report Configs", flush=True)
    legacy_report_configs = legacy_api.list_organisation_report_configs()
    c1_report_configs = c1_api.list_organisation_report_configs()

    cont_migration = check_existing_c1_report_configs(
        c1_api=c1_api,
        legacy_report_configs=legacy_report_configs,
        c1_report_configs=c1_report_configs,
        rconf_type=rconf_type,
    )
    if not cont_migration:
        return

    for report_config in legacy_report_configs:
        log.info(f"  --> Report Config: {report_config.title}")
        exec_migration_func(
            lambda: c1_api.create_organisation_report_config(
                report_conf=report_config.configuration
            )
        )


def create_user_defined_groups(
    legacy_api: LegacyConformityAPI, c1_api: CloudOneConformityAPI
):
    log.info("Creating groups")
    legacy_groups = legacy_api.list_groups(
        include_group_types=[Group.GROUP_TYPE_USER_DEFINED]
    )
    c1_groups = set(
        c1_api.list_groups(include_group_types=[Group.GROUP_TYPE_USER_DEFINED])
    )

    for group in legacy_groups:
        log.info(f" --> Group: {group.name}, Tags: {group.tags}", end="", flush=True)
        if group in c1_groups:
            log.info(" - Already exists! Skipping it.")
            continue
        exec_migration_func(
            lambda: c1_api.create_group(name=group.name, tags=group.tags)
        )

    if not legacy_groups:
        log.info(" --> No group found.")


def add_azure_group(mg: Group, c1_api: CloudOneConformityAPI):
    azure_conf = mg.cloud_data["azure"]  # type: ignore
    directory_name = mg.name
    directory_id = azure_conf["directoryId"]
    app_client_id = azure_conf["applicationId"]
    app_client_key = prompt_azure_app_client_id(
        directory_name, directory_id, app_client_id
    )
    c1_api.create_azure_directory(
        name=mg.name,
        directory_id=directory_id,
        app_client_id=app_client_id,
        app_client_key=app_client_key,
    )


def add_managed_groups(legacy_api: LegacyConformityAPI, c1_api: CloudOneConformityAPI):
    legacy_managed_groups = legacy_api.list_groups(
        include_group_types=[Group.GROUP_TYPE_MANAGED_GROUP]
    )
    c1_managed_groups = c1_api.list_groups(
        include_group_types=[Group.GROUP_TYPE_MANAGED_GROUP]
    )
    c1_managed_groups_set = set(c1_managed_groups)

    for mg in legacy_managed_groups:
        if mg in c1_managed_groups_set:
            # log.info(f"Managed Group {mg.name} ({mg.cloud_type.upper()}) already exists!")
            continue
        if mg.cloud_type == "azure":
            exec_migration_func(lambda: add_azure_group(mg=mg, c1_api=c1_api))


def prompt_azure_app_client_id(directory_name, directory_id, app_client_id) -> str:
    log.info(
        f"""
Please enter the App registration key for the following Active Directory:
If you lost the key, you may generate a new Client Secret on your Azure App Registration.
    Active Directory Name: {directory_name}
    Active Directory Tenant ID: {directory_id}
    App registration Application ID: {app_client_id}
"""
    )
    return ask_input("App registration key:", mask_input=True)


def _filter_users_with_verified_mobile(
    c1_user_ids: List[str],
    mobile_verified_c1_users: Set[str],
    c1_user_id_email_map: Dict[str, str],
) -> List[str]:

    c1_user_ids_set = set(c1_user_ids)
    unverified_user_ids = c1_user_ids_set.difference(mobile_verified_c1_users)
    for user_id in unverified_user_ids:
        email = c1_user_id_email_map.get(user_id, user_id)
        log.warn(
            f"User {email} doesn't have a mobile number verified. Excluding from SMS notification"
        )

    verified_c1_user_ids = c1_user_ids_set.intersection(mobile_verified_c1_users)
    return list(verified_c1_user_ids)


def _legacy_user_ids_to_c1_user_ids(
    legacy_user_ids: List[str],
    legacy_user_id_email_map: Dict[str, str],
    c1_email_user_id_map: Dict[str, str],
    channel: str,
) -> List[str]:
    c1_user_ids = []
    for legacy_user_id in legacy_user_ids:
        email = legacy_user_id_email_map.get(legacy_user_id)
        if email is None:
            log.warn(
                f"Cannot find email of Legacy Conformity user with an ID of: {legacy_user_id}. Excluding user from {channel} notification."
            )
            continue

        c1_user_id = c1_email_user_id_map.get(email)
        if c1_user_id is None:
            log.warn(
                f"Cannot find corresponding user in CloudOne Conformity: {email}. Excluding user from {channel} notification."
            )
            continue

        c1_user_ids.append(c1_user_id)

    return c1_user_ids


def copy_communication_channel_settings(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_acct_id: str,
    c1_acct_id: str,
    legacy_users: List[User],
    c1_users: List[User],
    c1_org_id: str,
):
    legacy_user_id_email_map = {user.user_id: user.email for user in legacy_users}
    c1_email_user_id_map = {user.email: user.user_id for user in c1_users}
    c1_user_id_email_map = {user.user_id: user.email for user in c1_users}
    mobile_verified_c1_users = {
        user.user_id for user in c1_users if user.is_mobile_verified
    }

    legacy_com_settings = legacy_api.get_communication_settings(acct_id=legacy_acct_id)
    candidate_com_settings: Set[CommunicationSettings] = set()
    for s in legacy_com_settings:
        legacy_conf = s.configuration
        c1_conf = legacy_conf
        if s.channel in ("email", "sms"):
            c1_conf["users"] = _legacy_user_ids_to_c1_user_ids(
                legacy_conf["users"],
                legacy_user_id_email_map,
                c1_email_user_id_map,
                s.channel,
            )
            if s.channel == "sms":
                c1_conf["users"] = _filter_users_with_verified_mobile(
                    c1_conf["users"], mobile_verified_c1_users, c1_user_id_email_map
                )

        candidate_com_settings.add(
            CommunicationSettings(
                com_setting_id=s.com_setting_id,
                channel=s.channel,
                enabled=s.enabled,
                filter=s.filter,
                configuration=c1_conf,
            )
        )

    c1_com_settings = set(c1_api.get_communication_settings(acct_id=c1_acct_id))
    new_com_settings = candidate_com_settings.difference(c1_com_settings)
    if new_com_settings:
        c1_api.create_communication_settings(
            com_settings=new_com_settings, acct_id=c1_acct_id, org_id=c1_org_id
        )


def migrate_account_configurations(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_acct_id: str,
    c1_acct_id: str,
    legacy_users: List[User],
    c1_users: List[User],
    c1_org_id: str,
):

    legacy_acct_details = legacy_api.get_account_details(acct_id=legacy_acct_id)
    name = legacy_acct_details.name
    environment = legacy_acct_details.environment
    cloud_type = legacy_acct_details.cloud_type
    env_suffix = acct_env_suffix(environment)
    log.info(
        f"Migrating account configurations for: {name}{env_suffix} [{cloud_type.upper()}]:"
    )

    log.info("  --> Updating account tags", flush=True)
    exec_migration_func(
        lambda: c1_api.update_account(
            acct_id=c1_acct_id,
            name=name,
            environment=environment,
            tags=legacy_acct_details.tags,
        )
    )

    log.info("  --> Copying account bot settings", flush=True)
    # bot_settings = legacy_api.get_account_bot_settings(acct_id=legacy_acct_id)
    bot_settings = legacy_acct_details.bot_settings
    if bot_settings:
        bot_settings.pop("lastModifiedFrom", None)
        bot_settings.pop("lastModifiedBy", None)
        if cloud_type == "aws" and str2bool(
            os.getenv("ENABLE_C1_AWS_CONFORMITY_BOT", "False")
        ):
            bot_settings["disabled"] = None
        exec_migration_func(
            lambda: c1_api.update_account_bot_settings(
                acct_id=c1_acct_id, settings=bot_settings
            )
        )

    log.info("  --> Copying account rules settings:", flush=True)
    exec_migration_func(
        lambda: copy_account_rules_settings(
            legacy_api=legacy_api,
            c1_api=c1_api,
            legacy_acct_id=legacy_acct_id,
            c1_acct_id=c1_acct_id,
            legacy_acct_details=legacy_acct_details,
            legacy_users=legacy_users,
        )
    )

    log.info("  --> Copying communication channel settings", flush=True)
    exec_migration_func(
        lambda: copy_communication_channel_settings(
            legacy_api=legacy_api,
            c1_api=c1_api,
            legacy_acct_id=legacy_acct_id,
            c1_acct_id=c1_acct_id,
            legacy_users=legacy_users,
            c1_users=c1_users,
            c1_org_id=c1_org_id,
        )
    )

    exec_migration_func(
        lambda: copy_account_report_configs(
            legacy_api=legacy_api,
            c1_api=c1_api,
            legacy_acct_id=legacy_acct_id,
            c1_acct_id=c1_acct_id,
        )
    )

    if has_suppressed_check(legacy_api=legacy_api, acct_id=legacy_acct_id):
        log.info("  --> Waiting for bot scan to finish ")
        exec_migration_func(
            lambda: wait_for_bot_scan_to_finish(c1_api=c1_api, acct_id=c1_acct_id)
        )

        log.info("  --> Copying suppressed checks")
        exec_migration_func(
            lambda: copy_suppressed_checks(
                legacy_api=legacy_api,
                c1_api=c1_api,
                legacy_acct_id=legacy_acct_id,
                c1_acct_id=c1_acct_id,
            )
        )
    else:
        log.info("  --> No suppressed check found to migrate")


def copy_account_rule_setting(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_acct_id: str,
    c1_acct_id: str,
    rule: Rule,
    user_map: dict,
):
    rule_id = rule.rule_id
    log.info(
        f"    --> Rule: {rule_id} ({'enabled' if rule.enabled else 'disabled'})",
        flush=True,
    )
    rule_with_notes = legacy_api.get_account_rule_setting(
        acct_id=legacy_acct_id, rule_id=rule_id, with_notes=True
    )

    note_msg = create_new_note_from_history_of_notes(
        notes=rule_with_notes.notes, user_map=user_map
    )

    c1_api.update_account_rule_setting(
        acct_id=c1_acct_id,
        rule_id=rule_id,
        setting=rule_with_notes.setting,
        note=note_msg,
    )


def copy_account_rules_settings(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_acct_id: str,
    c1_acct_id: str,
    legacy_acct_details: AccountDetails,
    legacy_users: List[User],
):

    user_map = {user.user_id: user for user in legacy_users}
    for rule in legacy_acct_details.rules:
        exec_migration_func(
            lambda: copy_account_rule_setting(
                legacy_api=legacy_api,
                c1_api=c1_api,
                legacy_acct_id=legacy_acct_id,
                c1_acct_id=c1_acct_id,
                rule=rule,
                user_map=user_map,
            )
        )


def truncate_txt_to_length(txt: str, length=-1, truncated_suffix="") -> str:
    if length == -1 or (0 <= len(txt) <= length):
        return txt

    return txt[: length - len(truncated_suffix)] + truncated_suffix


def get_most_recent_note_msg(notes: List[Note]) -> str:
    if not notes:
        return ""
    note = sorted(notes, key=lambda note: note.created_ts, reverse=True)[0]
    return note.note


def create_new_note_from_history_of_notes(
    notes: List[Note], user_map: Dict[str, User]
) -> str:
    note_msg = "[Copied settings via migration tool]"
    if not notes:
        return f"{note_msg} No history of notes found."

    note_frags = []
    sorted_notes = sorted(notes, key=lambda note: note.created_ts, reverse=True)
    for note in sorted_notes:
        user = user_map.get(note.created_by)
        user_name = f"{user.first_name} {user.last_name}" if user else ""
        ts = int(note.created_ts / 1000)
        dt_str = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat(sep=" ")
        note_frag = f"On: {dt_str}\nBy: {user_name}\nNote: {note.note}"
        note_frags.append(note_frag)

    note_history = "\n\n".join(note_frags)

    note_msg = f"""{note_msg} History of notes:
-----------------------
{note_history}
-----------------------
"""

    return note_msg


def exec_migration_func(migration_func: Callable) -> Any:
    if not str2bool(os.getenv("SKIP_MIGRATION_FAILURES", "False")):
        return migration_func()
    try:
        return migration_func()
    except Exception:
        log.exception("There was a failure in migration")


def wait_for_bot_scan_to_finish(c1_api: CloudOneConformityAPI, acct_id: str):
    while not c1_api.is_bot_scan_done(acct_id=acct_id):
        time.sleep(app_config()["BOT_SCAN_CHECK_INTERVAL_IN_SECS"])
        continue


def has_suppressed_check(legacy_api: LegacyConformityAPI, acct_id: str) -> bool:
    checks = legacy_api.get_suppressed_checks(acct_id=acct_id, limit=1)
    return len(list(checks)) > 0


def copy_suppressed_check(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    c1_acct_id: str,
    legacy_check: Check,
):
    log.info(
        f"    --> {legacy_check.rule_id}|{legacy_check.region}|{legacy_check.service}|{legacy_check.resource_name}|{legacy_check.resource}",
        flush=True,
    )
    filters: Dict[str, Any] = {
        "ruleIds": [legacy_check.rule_id],
        "services": [legacy_check.service],
        "regions": [legacy_check.region],
        "statuses": "FAILURE",
    }
    if legacy_check.resource:
        filters["resourceSearchMode"] = "text"
        filters["resource"] = legacy_check.resource
    c1_checks = list(c1_api.get_checks(acct_id=c1_acct_id, filters=filters))
    c1_checks_map = {c: c for c in c1_checks}
    c1_check = c1_checks_map.get(legacy_check)
    if c1_check is None:
        show_instructions_for_missing_check(legacy_check)
        return
    legacy_check_detail = legacy_api.get_check_detail(
        check_id=legacy_check.check_id, with_notes=True
    )
    note_msg = get_most_recent_note_msg(legacy_check_detail.notes)
    if not note_msg:
        note_msg = "[Migration tool: No note found from the source Check]"
    note_msg = truncate_txt_to_length(txt=note_msg, length=200, truncated_suffix="..")
    # log.info(f"Note: {note_msg}")
    c1_api.suppress_check(
        check_id=c1_check.check_id,
        suppressed_until=legacy_check.suppressed_until,
        note=note_msg,
    )


def copy_suppressed_checks(
    legacy_api: LegacyConformityAPI,
    c1_api: CloudOneConformityAPI,
    legacy_acct_id: str,
    c1_acct_id: str,
):
    legacy_checks = legacy_api.get_suppressed_checks(acct_id=legacy_acct_id)
    for legacy_check in legacy_checks:
        exec_migration_func(
            lambda: copy_suppressed_check(
                legacy_api=legacy_api,
                c1_api=c1_api,
                c1_acct_id=c1_acct_id,
                legacy_check=legacy_check,
            )
        )


def show_instructions_for_missing_check(check: Check):
    log.warn(
        f"""
    Can't find the corresponding check in Cloud One. Please manually suppress the check below or try running Conformity Bot
    on this account and run this tool again.
        RuleID: {check.rule_id}
        Region: {check.region}
        Resource: {check.resource}
        Message: {check.message}
"""
    )


def verify_users_mobile_numbers(users_to_verify_mobile: Iterable[User]):
    log.info(
        """
Please have the following users in your CloudOne Account to have their mobile
numbers verified. If they are one of the recipients for SMS notifications,
then it is important to do this now before we proceed with migration:
"""
    )
    for user in users_to_verify_mobile:
        log.info(
            f" --> {user.first_name} {user.last_name}; Email={user.email}; Mobile={user.mobile_number}"
        )
    log.info("")
    ask_when_mobile_verification__done()


def invite_users(users_to_invite: Iterable[User]):
    log.info(
        """
Please invite the following users to your CloudOne Account.
If they are one of the recipients for your communication channel,
then it is important to add them now before we proceed with migration:
"""
    )
    for user in users_to_invite:
        log.info(f" --> {user.first_name} {user.last_name}; Email={user.email}")
    log.info("")
    ask_when_user_invite_done()


def ask_confirmation_or_auto_overwrite(
    msg: str, default=False, ask_if_sure=False
) -> bool:
    if str2bool(os.getenv("C1_CONFORMITY_OVERWRITE_ALL", "False")):
        return True

    return ask_confirmation(msg=msg, default=default, ask_if_sure=ask_if_sure)


def ask_confirmation_with_text_verification(
    msg: str, verify_text: str, default=False
) -> bool:
    questions = [
        {
            "type": "confirm",
            "message": msg,
            "name": "continue",
            "default": default,
        },
    ]
    answer = prompt(questions=questions)
    cont = answer["continue"]
    if not cont:
        return False

    entered_text = input(
        f"Please enter the text '{verify_text}' to confirm this (exclude single quote): "
    )
    return entered_text == verify_text


def ask_confirmation(msg: str, default=False, ask_if_sure=False) -> bool:
    questions = [
        {
            "type": "confirm",
            "message": msg,
            "name": "continue",
            "default": default,
        },
    ]
    while True:
        answer = prompt(questions=questions)
        cont = answer["continue"]
        if not ask_if_sure or not cont:
            return cont

        sure = ask_confirmation(
            f"You chose {'Yes' if cont else 'No'}. Are you sure?", default=False
        )
        if sure:
            return cont


def ask_choices(msg: str, choices: List[str], default=1):
    questions = [
        {
            "type": "list",
            "message": msg,
            "name": "choice",
            "choices": choices,
            "default": default,
        },
    ]
    answer = prompt(questions=questions)
    return answer["choice"]


def ask_when_mobile_verification__done() -> None:
    while True:

        is_done = ask_confirmation(
            "Are you done with mobile verification?",
            default=False,
            ask_if_sure=True,
        )
        if is_done:
            return


def ask_when_user_invite_done() -> None:
    while True:

        is_done = ask_confirmation(
            "Are you done adding the users to Cloud One?",
            default=False,
            ask_if_sure=True,
        )
        if is_done:
            return


def ask_input(msg: str, mask_input=False) -> str:
    name = "input"
    questions = [
        {
            "type": "password" if mask_input else "input",
            "message": msg,
            "name": name,
            "default": "",
        },
    ]
    answer = prompt(questions=questions)
    return answer[name]


def pretty_print_com_settings(com_settings):
    for s in com_settings:
        log.debug(s)


@click.group(
    help=f"Conformity Migration Tool (ver {tool_version})\n\nMigrates your visiblity information in cloudconformity.com to cloudone.trendmicro.com",
)
@click.version_option(version=tool_version)
def cli():
    pass


@cli.command(help="Configures migration tool")
def configure():
    create_user_config(user_config_path())


@cli.command(help="Runs migration")
@click.option(
    "--skip-aws-prompt",
    is_flag=True,
    envvar="SKIP_AWS_PROMPT",
    show_envvar=True,
    required=False,
    default=False,
    help="Skips prompting for manually editing AWS Conformity stack",
)
@click.option(
    "--overwrite-all",
    is_flag=True,
    envvar="C1_CONFORMITY_OVERWRITE_ALL",
    show_envvar=True,
    required=False,
    default=False,
    help="Always overwrites Cloud One Conformity configurations. This prevents asking user for confirmation.",
)
@click.option(
    "--skip-migration-failures",
    is_flag=True,
    envvar="SKIP_MIGRATION_FAILURES",
    show_envvar=True,
    required=False,
    default=False,
    help="Will always allow migration to continue even when some configuration fails to migrate.",
)
@click.option(
    "--include-accounts-file",
    required=False,
    type=str,
    help="CSV file containing accounts that will be the only ones included in migration. Each row should consists of 2 fields: first is the account name and second is the environment as they appear on Conformity Dashboard. An empty file means the tool won't include any account in the migration.",
)
@click.option(
    "--exclude-accounts-file",
    required=False,
    type=str,
    help="CSV file containing accounts that will be excluded from migration. Each row should consists of 2 fields: first is the account name and second is the environment as they appear on Conformity Dashboard.",
)
@click.option(
    "--enable-aws-bot",
    is_flag=True,
    envvar="ENABLE_C1_AWS_CONFORMITY_BOT",
    show_envvar=True,
    required=False,
    default=False,
    help="Enables bot settings for all migrated AWS accounts on Cloud One Conformity.",
)
def run(
    skip_aws_prompt: bool,
    overwrite_all: bool,
    skip_migration_failures: bool,
    include_accounts_file: str,
    exclude_accounts_file: str,
    enable_aws_bot: bool,
):
    include_accts: Optional[Set[AccountEnv]] = None
    exclude_accts: Optional[Set[AccountEnv]] = None
    if include_accounts_file:
        include_accts = read_accts_file(accounts_file=include_accounts_file)
    if exclude_accounts_file:
        exclude_accts = read_accts_file(accounts_file=exclude_accounts_file)

    os.environ["SKIP_AWS_PROMPT"] = "True" if skip_aws_prompt else "False"
    os.environ["C1_CONFORMITY_OVERWRITE_ALL"] = "True" if overwrite_all else "False"
    os.environ["SKIP_MIGRATION_FAILURES"] = (
        "True" if skip_migration_failures else "False"
    )
    os.environ["ENABLE_C1_AWS_CONFORMITY_BOT"] = "True" if enable_aws_bot else "False"
    try:
        run_migration(
            legacy_api=legacy_conformity_api(),
            c1_api=c1_conformity_api(),
            include_accts=include_accts,
            exclude_accts=exclude_accts,
        )
    except ConformityError as e:
        log.error(e)
        log.error(e.details)
        # raise e


def read_accts_file(accounts_file: str) -> Set[AccountEnv]:
    accts = set()
    accounts_path = Path(accounts_file.strip())
    if not accounts_path.exists():
        raise FileNotFoundError(f"File does not exists: {accounts_file}")
    if not accounts_path.is_file():
        raise FileNotFoundError(f"Not a regular file: {accounts_file}")

    with open(accounts_path, newline="", mode="r") as fh:
        csvr = csv.reader(fh, dialect="excel")
        for row in csvr:
            acct_name = row[0]
            acct_env = row[1] if len(row) > 1 else ""
            accts.add((acct_name, acct_env))
    return accts


@cli.command(
    "empty-legacy",
    help="Deletes all accounts and configurations in Legacy Conformity",
)
def empty_legacy():
    try:
        empty_legacy_conformity(legacy_api=legacy_conformity_api())
    except ConformityError as e:
        log.error(e)
        log.error(e.details)
        # raise e


@cli.command(
    "empty-c1",
    help="Deletes all accounts and configurations in Cloud One Conformity",
    hidden=True,
)
def empty_c1():
    try:
        empty_c1_conformity(c1_api=c1_conformity_api())
    except ConformityError as e:
        log.error(e)
        log.error(e.details)
        # raise e


if __name__ == "__main__":
    cli()



import logging
import os
import random
import sys
from functools import lru_cache
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, Dict

import yaml
from requests import Response, Session
from requests.adapters import BaseAdapter, HTTPAdapter
from urllib3 import Retry
from vcr import VCR

from conformity_migration.conformity_api import (
    CloudOneConformityAPI,
    DefaultConformityAPI,
    LegacyConformityAPI,
    WorkaroundFixConformityAPI,
)

from .logger import (
    AppLogger,
    Logger,
    NoStrackTraceExceptionFormatter,
    WithStrackTraceExceptionFormatter,
)
from .utils import str2bool

script_dirpath = Path(__file__).parent

USER_CONF_FILENAME = "user_config.yml"
APP_CONF_FILENAME = "config.yml"


def _load_yaml_config(yaml_path: Path):
    with open(yaml_path, mode="r") as fh:
        return yaml.load(fh, Loader=yaml.SafeLoader)


def app_config_path() -> Path:
    return script_dirpath.joinpath(APP_CONF_FILENAME)


@lru_cache(maxsize=1)
def app_config() -> Dict[str, Any]:
    return _load_yaml_config(app_config_path())


def ask_user_to_run_configure():
    print(
        "Please configure migration tool by running this command: conformity-migration configure"
    )
    sys.exit(1)


def user_config_path() -> Path:
    return Path(USER_CONF_FILENAME)


@lru_cache(maxsize=1)
def user_config() -> Dict[str, Any]:
    path = user_config_path()
    if not path.exists():
        ask_user_to_run_configure()
    return _load_yaml_config(path)


class LoggerHTTPAdapter(BaseAdapter):
    def __init__(self, adapter: BaseAdapter, logger=Logger):
        self._adapter = adapter
        self._log = logger

    def _build_req_resp_txt(self, resp: Response):
        req = resp.request
        req_body = str(req.body) if req.body else ""
        return f"""[Request]
{req.method} {req.url}

{req_body}
[Response]
{resp.status_code} {resp.reason}
Content-Type: {resp.headers['Content-Type']}

{resp.text}"""

    def send(self, *args, **kwargs) -> Response:
        resp = self._adapter.send(*args, **kwargs)
        if not resp.ok:
            req_resp = self._build_req_resp_txt(resp)
            self._log.error(req_resp, file_only=True)
        return resp

    def close(self) -> None:
        return self._adapter.close()


class FakeErrorHTTPAdapter(BaseAdapter):
    def __init__(self, adapter: BaseAdapter):
        self._adapter = adapter

    def _fake_error(self, resp: Response):
        req = resp.request
        if req.method == "GET":
            return
        is_error = random.choice((True, False))
        if is_error:
            resp.status_code = 499
            resp.reason = "Fake Error :-)"

    def send(self, *args, **kwargs) -> Response:
        resp = self._adapter.send(*args, **kwargs)
        self._fake_error(resp)
        return resp

    def close(self) -> None:
        return self._adapter.close()


class TimeoutHTTPAdapter(BaseAdapter):
    def __init__(self, adapter: BaseAdapter, conn_timeout: float, read_timeout: float):
        self._adapter = adapter
        self._conn_timeout = conn_timeout
        self._read_timeout = read_timeout

    def send(self, *args, **kwargs) -> Response:
        timeout = (self._conn_timeout, self._read_timeout)
        kwargs["timeout"] = timeout
        return self._adapter.send(*args, **kwargs)

    def close(self) -> None:
        return self._adapter.close()


class VcrHTTPAdapter(BaseAdapter):
    def __init__(self, adapter: BaseAdapter, vcr: VCR, vcr_file: str):
        self._adapter = adapter
        self._vcr = vcr
        self._vcr_file = vcr_file

    def send(self, *args, **kwargs) -> Response:
        with self._vcr.use_cassette(path=self._vcr_file):
            return self._adapter.send(*args, **kwargs)

    def close(self) -> None:
        return self._adapter.close()


def _http_adapter() -> BaseAdapter:
    app_conf = app_config()

    adapter = HTTPAdapter(
        max_retries=Retry(
            total=None,
            connect=0,
            read=0,
            redirect=0,
            other=0,
            backoff_factor=app_conf["API_RETRY_BACKOFF_FACTOR"],
            status=app_conf["API_RETRY_COUNT"],  # max retries for any status_forcelist
            status_forcelist=app_conf["API_RETRY_HTTP_STATUSES"],
            allowed_methods=False,  # false means retry on all Methods
            respect_retry_after_header=True,
        )
    )

    adapter = TimeoutHTTPAdapter(
        adapter,
        conn_timeout=app_conf["API_CONNECTION_TIMEOUT"],
        read_timeout=app_conf["API_READ_TIMEOUT"],
    )

    adapter = LoggerHTTPAdapter(adapter=adapter, logger=logger())

    return adapter


def _vcr_adapter(
    adapter: BaseAdapter,
    vcr_file: str,
    vcr_mode: str,
    fake_api_key: str,
    decode_compressed_response=False,
) -> BaseAdapter:
    if vcr_file:
        vcr = VCR(
            record_mode=vcr_mode,
            match_on=("method", "scheme", "host", "port", "path", "query", "body"),
            filter_headers=[("Authorization", f"ApiKey {fake_api_key}")],
            decode_compressed_response=True,
        )
        adapter = VcrHTTPAdapter(adapter=adapter, vcr=vcr, vcr_file=vcr_file)
    return adapter


def _legacy_http() -> Session:
    sess = Session()
    adapter = _http_adapter()
    adapter = _vcr_adapter(
        adapter=adapter,
        vcr_file=os.getenv("LEG_VCR_FILE", ""),
        vcr_mode=os.getenv("LEG_VCR_MODE", "none"),
        fake_api_key="fake-api-key-for-legacy_conformity",
    )
    sess.mount("https://", adapter=adapter)
    return sess


def _c1_http() -> Session:
    sess = Session()
    adapter = _http_adapter()
    adapter = _vcr_adapter(
        adapter=adapter,
        vcr_file=os.getenv("C1_VCR_FILE", ""),
        vcr_mode=os.getenv("C1_VCR_MODE", "none"),
        fake_api_key="fake-api-key-for-c1_conformity",
    )
    if str2bool(os.getenv("FAKE_C1_HTTP_ERROR", "False")):
        adapter = FakeErrorHTTPAdapter(adapter=adapter)
    sess.mount("https://", adapter=adapter)
    return sess


def legacy_conformity_api() -> LegacyConformityAPI:
    user_conf = user_config()
    api_key = user_conf["LEGACY_CONFORMITY"]["API_KEY"]
    base_url = user_conf["LEGACY_CONFORMITY"]["API_BASE_URL"]

    api = DefaultConformityAPI(api_key=api_key, base_url=base_url, http=_legacy_http())
    api = WorkaroundFixConformityAPI(api)
    api = LegacyConformityAPI(api)
    return api


def c1_conformity_api() -> CloudOneConformityAPI:
    user_conf = user_config()
    api_key = user_conf["CLOUD_ONE_CONFORMITY"]["API_KEY"]
    base_url = user_conf["CLOUD_ONE_CONFORMITY"]["API_BASE_URL"]

    api = DefaultConformityAPI(api_key=api_key, base_url=base_url, http=_c1_http())
    api = WorkaroundFixConformityAPI(api)
    api = CloudOneConformityAPI(api)
    return api


@lru_cache(maxsize=1)
def logger() -> Logger:
    logger = logging.getLogger("app")
    logger.setLevel(logging.INFO)

    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.INFO)
    ch_fmt = NoStrackTraceExceptionFormatter(fmt="%(message)s")
    ch.setFormatter(ch_fmt)
    ch.addFilter(lambda logrec: not logrec.file_only)  # type: ignore
    logger.addHandler(ch)

    fh_fmt = WithStrackTraceExceptionFormatter(
        fmt="[%(asctime)s] %(levelname)s %(message)s"
    )

    log_fh = RotatingFileHandler(
        filename="conformity-migration.log", maxBytes=1024**2, backupCount=4
    )
    log_fh.setLevel(logging.DEBUG)
    log_fh.setFormatter(fmt=fh_fmt)
    logger.addHandler(log_fh)

    err_fh = RotatingFileHandler(
        filename="conformity-migration-error.log", maxBytes=1024**2, backupCount=4
    )
    err_fh.setLevel(logging.ERROR)
    err_fh.setFormatter(fmt=fh_fmt)
    logger.addHandler(err_fh)

    log_backoff = app_config()["LOG_BACKOFF"]
    if log_backoff:
        backoff_logger = logging.getLogger("backoff")
        backoff_logger.addHandler(log_fh)
        backoff_logger.addHandler(err_fh)

    return AppLogger(logger=logger)



import json
from typing import Any, Dict, Iterable, List, Optional
from urllib.parse import quote

import backoff
import requests

from .models import (
    Account,
    AccountDetails,
    Check,
    CommunicationSettings,
    Group,
    Note,
    Profile,
    ReportConfig,
    Rule,
    User,
)
from .typing import Protocol


class ConformityError(Exception):
    def __init__(self, *args: object, details="") -> None:
        super().__init__(*args)
        self.details = details


class ConformityConnectionError(ConformityError):
    pass


class ConformityUnauthorizedError(ConformityError):
    pass


class ConformityForbiddenError(ConformityError):
    pass


class ConformityResourceNotFoundError(ConformityError):
    pass


class ConformityClientError(ConformityError):
    pass


class ConformityServerInternalError(ConformityError):
    pass


class ConformityOtherError(ConformityError):
    pass


class ConformityAPI(Protocol):
    def current_user(self) -> User:
        pass

    def get_organisation_external_id(self) -> str:
        pass

    def get_account_access_configuration(self, acct_id: str) -> Dict[str, Any]:
        pass

    def add_aws_account(
        self,
        name: str,
        environment: str,
        role_arn: str,
        external_id: str,
        subscription_type="advanced",
    ) -> Dict[str, Any]:
        pass

    def add_azure_subscription(
        self,
        name: str,
        environment: str,
        subscription_id: str,
        active_directory_id: str,
    ) -> Dict[str, Any]:
        pass

    def get_group_details(self, group_id: str) -> Dict[str, Any]:
        pass

    def get_organisation_id(self) -> str:
        pass

    def get_all_users(self) -> List[User]:
        pass

    def list_groups(self, include_group_types: List[str] = None) -> List[Group]:
        pass

    def delete_account(self, acct_id: str) -> dict:
        pass

    def list_accounts(self) -> List[Account]:
        pass

    def get_organisation_profile(self, include_rule_settings=False) -> Profile:
        pass

    def update_organisation_profile(self, profile: Profile) -> Profile:
        pass

    def reset_organisation_profile(self) -> dict:
        pass

    def get_custom_profiles(self) -> List[Profile]:
        pass

    def get_profile(self, profile_id: str, include_rule_settings=False) -> Profile:
        pass

    def create_new_profile(self, profile: Profile) -> Profile:
        pass

    def delete_profile(self, profile_id: str):
        pass

    def list_organisation_report_configs(self) -> List[ReportConfig]:
        pass

    def list_group_report_configs(self, group_id: str) -> List[ReportConfig]:
        pass

    def list_account_report_configs(self, acct_id: str) -> List[ReportConfig]:
        pass

    def create_organisation_report_config(self, report_conf: Dict[str, Any]):
        pass

    def create_group_report_config(self, report_conf: Dict[str, Any], group_id: str):
        pass

    def create_account_report_config(self, report_conf: Dict[str, Any], acct_id: str):
        pass

    def delete_report_config(self, report_conf_id: str):
        pass

    def delete_group(self, group_id: str) -> dict:
        pass

    def create_group(self, name, tags=List[str]):
        pass

    def create_azure_directory(
        self, name: str, directory_id: str, app_client_id: str, app_client_key: str
    ):
        pass

    def delete_communication_settings(self, com_setting_id: str) -> dict:
        pass

    def get_communication_settings(self, acct_id: str) -> List[CommunicationSettings]:
        pass

    def create_communication_settings(
        self, com_settings: Iterable[CommunicationSettings], acct_id: str, org_id: str
    ):
        pass

    def get_account_details(self, acct_id: str) -> AccountDetails:
        pass

    def update_account(
        self, acct_id: str, name: str, environment: str, tags: List[str]
    ):
        pass

    def update_account_bot_settings(self, acct_id: str, settings: dict):
        pass

    def get_account_rule_setting(
        self, acct_id: str, rule_id: str, with_notes=False
    ) -> Rule:
        pass

    def update_account_rule_setting(
        self, acct_id: str, rule_id: str, setting: dict, note: str = "Copied from API"
    ):
        pass

    def is_bot_scan_done(self, acct_id: str) -> bool:
        pass

    def get_suppressed_checks(self, acct_id: str, limit=0) -> Iterable[Check]:
        pass

    def get_checks(
        self, acct_id: str, filters: Optional[Dict[str, Any]] = None, limit=0
    ) -> Iterable[Check]:
        pass

    def get_check_detail(
        self, check_id: str, with_notes=False, notes_limit=100
    ) -> Check:
        pass

    def suppress_check(
        self, check_id: str, suppressed_until: Optional[int], note="Copied from API"
    ):
        pass


class DefaultConformityAPI:
    def __init__(
        self, api_key: str, base_url: str, http: requests.Session = None
    ) -> None:
        self._api_key = api_key
        self._base_url = base_url.strip().rstrip("/")
        self.http = requests.Session() if http is None else http
        self._headers = {
            "Authorization": f"ApiKey {self._api_key}",
            "Content-Type": "application/vnd.api+json",
        }
        self._current_user: Optional[User] = None
        self._validate_api()
        self._organisation_external_id = ""

    def _err_details(self, resp: requests.Response) -> str:
        req = resp.request
        return f"""Request:
{req.method} {req.url}
Response:
{resp.status_code} {resp.reason}
{resp.text}"""

    def _raise_for_status(self, resp: requests.Response):
        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            err_resp: requests.Response = e.response
            status_code = err_resp.status_code
            details = self._err_details(resp=err_resp)
            msg = str(e)
            if status_code == 401:
                raise ConformityUnauthorizedError(msg, details=details) from e
            elif status_code == 403:
                raise ConformityForbiddenError(msg, details=details) from e
            elif status_code == 404:
                raise ConformityResourceNotFoundError(msg, details=details) from e
            elif 400 <= status_code < 500:
                raise ConformityClientError(msg, details=details)
            elif 500 <= status_code < 600:
                raise ConformityServerInternalError(msg, details=details) from e
            else:
                raise ConformityOtherError(msg, details=details) from e

    def _get_request(self, url, params=None):
        return self._exec_request("GET", url, params=params)

    def _post_request(self, url, data):
        return self._exec_request("POST", url, data=data)

    def _delete_request(self, url):
        return self._exec_request("DELETE", url)

    def _patch_request(self, url, data):
        return self._exec_request("PATCH", url, data=data)

    def _exec_request(self, method, url, params=None, data=None):
        json_data = json.dumps(data, indent=4) if data else None
        try:
            resp = self.http.request(
                method=method,
                url=url,
                params=params,
                data=json_data,
                headers=self._headers,
            )
        except requests.exceptions.ConnectTimeout as e:
            raise ConformityConnectionError(
                f"Cannot connect to {self._base_url}"
            ) from e
        else:
            self._raise_for_status(resp)

            return resp.json()

    def _validate_api(self):
        try:
            user = self.current_user()
        except ConformityForbiddenError as e:
            raise ConformityForbiddenError(
                f"API Key does not have permission for: {self._base_url}"
            ) from e
        else:
            if user.role != User.ROLE_ADMIN:
                raise ConformityForbiddenError(
                    f"Insufficient permisison. API Key must have an ADMIN privilege for: {self._base_url}"
                )

    def current_user(self) -> User:
        if not self._current_user:
            res = self._get_request(f"{self._base_url}/users/whoami")
            user = self._user_dict_to_user_obj(res["data"])
            self._current_user = user
        return self._current_user

    def delete_account(self, acct_id: str) -> dict:
        res = self._delete_request(f"{self._base_url}/accounts/{acct_id}")
        return res

    def list_accounts(self) -> List[Account]:
        res = self._get_request(f"{self._base_url}/accounts")
        return [Account(acct_data=acct_data) for acct_data in res["data"]]

    def get_organisation_external_id(self) -> str:
        if not self._organisation_external_id:
            res = self._get_request(f"{self._base_url}/organisation/external-id")
            self._organisation_external_id = res["data"]["id"]
        return self._organisation_external_id

    def get_account_access_configuration(self, acct_id) -> dict:
        res = self._get_request(f"{self._base_url}/accounts/{acct_id}/access")
        return res["attributes"]["configuration"]

    @backoff.on_exception(
        wait_gen=backoff.expo,
        factor=2,
        exception=ConformityClientError,
        max_time=30,
        jitter=None,
    )
    def add_aws_account(
        self,
        name: str,
        environment: str,
        role_arn: str,
        external_id: str,
        subscription_type="advanced",
    ) -> dict:

        res = self._post_request(
            url=f"{self._base_url}/accounts",
            data={
                "data": {
                    "attributes": {
                        "name": name,
                        "environment": environment,
                        "access": {
                            "keys": {"roleArn": role_arn, "externalId": external_id}
                        },
                        "subscriptionType": subscription_type,
                    },
                }
            },
        )

        return res["data"]

    def add_azure_subscription(
        self,
        name: str,
        environment: str,
        subscription_id: str,
        active_directory_id: str,
    ) -> dict:

        res = self._post_request(
            url=f"{self._base_url}/accounts/azure",
            data={
                "data": {
                    "attributes": {
                        "name": name,
                        "environment": environment,
                        "access": {
                            "subscriptionId": subscription_id,
                            "activeDirectoryId": active_directory_id,
                        },
                    },
                }
            },
        )

        return res["data"]

    def update_account(
        self, acct_id: str, name: str, environment: str, tags: List[str]
    ):
        res = self._patch_request(
            url=f"{self._base_url}/accounts/{acct_id}",
            data={
                "data": {
                    "attributes": {
                        "name": name,
                        "environment": environment,
                        "tags": tags,
                    },
                },
            },
        )
        return res

    def get_account_bot_settings(self, acct_id: str) -> dict:
        res = self._get_request(f"{self._base_url}/accounts/{acct_id}/settings/bot")
        return res["data"]["attributes"]["settings"]["bot"]

    def update_account_bot_settings(self, acct_id: str, settings: dict):
        res = self._patch_request(
            url=f"{self._base_url}/accounts/{acct_id}/settings/bot",
            data={
                "data": {"attributes": {"settings": {"bot": settings}}},
            },
        )
        return res

    def get_account_details(self, acct_id: str) -> AccountDetails:
        res = self._get_request(f"{self._base_url}/accounts/{acct_id}")
        return AccountDetails(acct_data=res["data"])

    def get_account_rules_settings(self, acct_id: str) -> list:
        try:
            res = self._get_request(
                f"{self._base_url}/accounts/{acct_id}/settings/rules"
            )
            return res["data"]["attributes"]["settings"]["rules"]
        except ConformityResourceNotFoundError:
            print("ResourceNotFoundError!!! So returning empty list instead :-)")
            return []
        except Exception as e:
            raise e

    def update_account_rule_settings(
        self, acct_id: str, settings: dict, note: str = "Copied from API"
    ):
        res = self._patch_request(
            url=f"{self._base_url}/accounts/{acct_id}/settings/rules",
            data={
                "data": {"attributes": {"note": note, "ruleSettings": settings}},
            },
        )
        return res

    def get_account_rule_setting(
        self, acct_id: str, rule_id: str, with_notes=False
    ) -> Rule:
        res = self._get_request(
            f"{self._base_url}/accounts/{acct_id}/settings/rules/{rule_id}",
            params={"notes": "true" if with_notes else "false"},
        )
        # print(json.dumps(res, indent=4))
        setting = res["data"]["attributes"]["settings"]["rules"][0]
        notes: List[Note] = []
        if with_notes:
            meta = res["meta"]
            ns = meta.get("notes")
            if ns is None:
                ns = meta["deprecation"]["notes"]
            notes = [
                Note(
                    note=n["note"],
                    created_by=n["createdBy"],
                    created_ts=n["createdDate"],
                )
                for n in ns
            ]
        return Rule(setting=setting, notes=notes)

    def update_account_rule_setting(
        self, acct_id: str, rule_id: str, setting: dict, note: str = "Copied from API"
    ):
        res = self._patch_request(
            url=f"{self._base_url}/accounts/{acct_id}/settings/rules/{rule_id}",
            data={
                "data": {"attributes": {"ruleSetting": setting, "note": note}},
            },
        )
        return res["data"]

    def get_group_details(self, group_id: str) -> dict:
        res = self._get_request(f"{self._base_url}/groups/{group_id}")
        return res["data"][0]

    def list_groups(self, include_group_types: List[str] = None) -> List[Group]:
        res = self._get_request(f"{self._base_url}/groups")
        if include_group_types is None:
            include_group_types = []
        groups = []
        for g in res["data"]:
            gattrib = g["attributes"]
            group_type = gattrib.get("group-type", Group.GROUP_TYPE_USER_DEFINED)
            if include_group_types and group_type not in include_group_types:
                continue
            group = Group(
                group_id=g["id"],
                name=gattrib["name"],
                tags=gattrib.get("tags"),
                group_type=group_type,
                cloud_type=gattrib.get("cloud-type"),
                cloud_data=gattrib.get("cloud-data"),
            )
            groups.append(group)
        return groups

    def create_group(self, name, tags=List[str]):
        res = self._post_request(
            url=f"{self._base_url}/groups",
            data={"data": {"attributes": {"name": name, "tags": tags}}},
        )
        return res["data"]

    def delete_group(self, group_id: str) -> dict:
        res = self._delete_request(f"{self._base_url}/groups/{group_id}")
        return res

    def get_organisation_id(self) -> str:
        res = self._get_request(f"{self._base_url}/users")
        return res["data"][0]["relationships"]["organisation"]["data"]["id"]

    def list_all_users(self) -> List[dict]:
        res = self._get_request(f"{self._base_url}/users")
        return res["data"]

    def _user_dict_to_user_obj(self, u: dict) -> User:
        user_attrib = u["attributes"]
        email: str = user_attrib.get("email", "")
        return User(
            user_id=u["id"],
            email=email,
            first_name=user_attrib.get("first-name", ""),
            last_name=user_attrib.get("last-name", ""),
            role=user_attrib["role"],
            mobile_number=user_attrib.get("mobile", ""),
            is_mobile_verified=user_attrib.get("mobile-verified", False),
            is_cloud_one_user=user_attrib.get("is-cloud-one-user", False),
        )

    def get_all_users(self) -> List[User]:
        res = self._get_request(f"{self._base_url}/users")
        users = []
        for u in res["data"]:
            user = self._user_dict_to_user_obj(u)
            if not user.email:  # skip users who does not have email, e.g. Api key user
                continue
            users.append(user)
        return users

    def get_user_details(self, user_id: str) -> dict:
        res = self._get_request(f"{self._base_url}/users/{user_id}")
        return res["data"]

    def invite_user(self, user: dict) -> List[dict]:
        attrib = user["attributes"]
        first_name = attrib["first-name"]
        last_name = attrib["last-name"]
        email = attrib["email"]
        role = attrib["role"]

        data = {
            "data": {
                "attributes": {
                    "firstName": first_name,
                    "lastName": last_name,
                    "email": email,
                    "role": role,
                }
            }
        }
        if role == "USER":
            data["data"]["attributes"]["accessList"] = user["relationships"][
                "accountAccessList"
            ]

        res = self._post_request(
            url=f"{self._base_url}/users",
            data=data,
        )
        return res["data"]

    def delete_communication_settings(self, com_setting_id: str) -> dict:
        res = self._delete_request(f"{self._base_url}/settings/{com_setting_id}")
        return res

    def get_communication_settings(self, acct_id: str) -> List[CommunicationSettings]:
        if acct_id:
            params = {"accountId": acct_id}
        else:
            params = {"includeParents": "true"}
        res = self._get_request(
            f"{self._base_url}/settings/communication", params=params
        )
        # return res["data"]
        settings: List[CommunicationSettings] = []
        for s in res["data"]:
            attrib = s["attributes"]
            settings.append(
                CommunicationSettings(
                    com_setting_id=s["id"],
                    channel=attrib["channel"],
                    enabled=attrib["enabled"],
                    filter=attrib.get("filter"),
                    configuration=attrib.get("configuration"),
                )
            )
        return settings

    def create_communication_settings(
        self, com_settings: Iterable[CommunicationSettings], acct_id: str, org_id: str
    ):
        # for cs in com_settings:
        #     print(cs)
        settings = []
        for cs in com_settings:
            if not cs.configuration:
                continue
            s: dict = {
                "type": "settings",
                "attributes": {
                    "type": "communication",
                    "enabled": cs.enabled,
                    "channel": cs.channel,
                    "filter": cs.filter,
                    "configuration": cs.configuration,
                },
                "relationships": {
                    "account": {
                        "data": None,
                    },
                    "organisation": {
                        "data": {
                            "type": "organisations",
                            "id": org_id,
                        }
                    },
                },
            }
            if acct_id:
                s["relationships"]["account"]["data"] = {
                    "type": "accounts",
                    "id": acct_id,
                }

            settings.append(s)

        res = self._post_request(
            url=f"{self._base_url}/settings/communication", data={"data": settings}
        )
        return res

    def create_azure_directory(
        self, name: str, directory_id: str, app_client_id: str, app_client_key: str
    ):
        res = self._post_request(
            url=f"{self._base_url}/azure/active-directories",
            data={
                "data": {
                    "attributes": {
                        "name": name,
                        "directoryId": directory_id,
                        "applicationId": app_client_id,
                        "applicationKey": app_client_key,
                    }
                }
            },
        )
        return res["data"]

    def _limit_reached(self, limit, total) -> bool:
        is_limited: bool = limit > 0
        limit_reached = is_limited and total >= limit
        # if limit_reached:
        #     print("Limit reached!")
        return limit_reached

    def _check_dict_to_check_obj(self, c: dict) -> Check:
        attrib: dict = c["attributes"]
        ns = attrib.get("notes", [])
        notes = [
            Note(
                note=n["note"],
                created_by=n["createdBy"],
                created_ts=n["created-date"],
            )
            for n in ns
        ]
        return Check(
            check_id=c["id"],
            acct_id=c["relationships"]["account"]["data"]["id"],
            rule_id=c["relationships"]["rule"]["data"]["id"],
            service=attrib["service"],
            region=attrib["region"],
            resource_name=attrib.get("resourceName", ""),
            resource=attrib.get("resource", ""),
            message=attrib["message"],
            suppressed=attrib.get("suppressed"),
            suppressed_until=attrib.get("suppressed-until"),
            notes=notes,
        )

    def get_checks(
        self, acct_id: str, filters: Optional[Dict[str, Any]] = None, limit=0
    ) -> Iterable[Check]:

        page_size_max = 100
        page_size = limit if (0 < limit < page_size_max) else page_size_max

        params = {
            "accountIds": acct_id,
            "page[size]": page_size,
        }
        if filters:
            for filter_name, filter_val in filters.items():
                params[f"filter[{filter_name}]"] = filter_val

        total_items = 0
        page_num = 0
        while True:
            params["page[number]"] = page_num
            res = self._get_request(
                f"{self._base_url}/checks",
                params=params,
            )
            data = res["data"]
            for c in data:
                yield self._check_dict_to_check_obj(c)
                total_items += 1
                if self._limit_reached(limit, total_items):
                    break

            meta = res["meta"]
            # print(meta)
            # print(f"total_items: {total_items}")
            if self._limit_reached(limit, total_items):
                break

            if total_items >= meta["total"]:
                break
            page_num += 1

    def get_suppressed_checks(self, acct_id: str, limit=0) -> Iterable[Check]:
        return self.get_checks(
            acct_id=acct_id,
            filters={"suppressed": True, "suppressedFilterMode": "v2"},
            limit=limit,
        )

    def suppress_check(
        self, check_id: str, suppressed_until: Optional[int], note="Copied from API"
    ):
        res = self._patch_request(
            url=f"{self._base_url}/checks/{quote(check_id, safe='')}",
            data={
                "data": {
                    "type": "checks",
                    "attributes": {
                        "suppressed": True,
                        "suppressed-until": suppressed_until,
                    },
                },
                "meta": {
                    "note": note,
                },
            },
        )
        return res["data"]

    def get_check_detail(
        self, check_id: str, with_notes=False, notes_limit=100
    ) -> Check:
        notes_limit = notes_limit if (0 < notes_limit < 100) else 100
        params = {"filter[notes]": "true" if with_notes else "false"}
        if with_notes:
            params["filter[notesLength]"] = notes_limit
        res = self._get_request(
            f"{self._base_url}/checks/{quote(check_id, safe='')}", params=params
        )
        return self._check_dict_to_check_obj(res["data"])

    def is_bot_scan_done(self, acct_id: str) -> bool:
        bot_status = self.get_account_details(acct_id=acct_id).bot_status
        return bot_status is None

    def get_custom_profiles(self) -> List[Profile]:
        res = self._get_request(f"{self._base_url}/profiles")
        profiles: List[Profile] = []
        for prof_data in res["data"]:
            settings = {"data": prof_data}
            profiles.append(Profile(settings=settings))
        return profiles

    def get_profile(self, profile_id: str, include_rule_settings=False) -> Profile:
        params = {"includes": "ruleSettings"} if include_rule_settings else None
        res = self._get_request(
            f"{self._base_url}/profiles/{profile_id}", params=params
        )
        return Profile(settings=res)

    def _get_org_profile_id(self) -> str:
        return f"organisation-{self.get_organisation_id()}"

    def get_organisation_profile(self, include_rule_settings=False) -> Profile:
        return self.get_profile(
            profile_id=self._get_org_profile_id(),
            include_rule_settings=include_rule_settings,
        )

    def update_organisation_profile(self, profile: Profile) -> Profile:
        profile.delete_meta()
        profile.delete_profile_id()
        profile_id = self._get_org_profile_id()
        profile.profile_id = profile_id
        res = self._patch_request(
            url=f"{self._base_url}/profiles/{profile_id}", data=profile.settings
        )
        return Profile(settings=res)

    @classmethod
    def create_empty_organisation_profile(
        cls, org_id: str, has_empty_included_field=True
    ) -> Profile:
        settings = {
            "meta": {},
            "data": {
                "type": "profiles",
                "id": f"organisation-{org_id}",
                "attributes": {
                    "name": "Organisational Profile",
                    "description": "Organisational Profile",
                },
                "relationships": {"ruleSettings": {"data": []}},
            },
        }
        if has_empty_included_field:
            settings["included"] = []
        return Profile(settings=settings)

    def reset_organisation_profile(self) -> dict:
        org_id = self.get_organisation_id()
        profile = self.create_empty_organisation_profile(
            org_id=org_id, has_empty_included_field=False
        )
        # print(json.dumps(profile.settings, indent=4))
        res = self._post_request(
            url=f"{self._base_url}/profiles", data=profile.settings
        )
        return res

    def create_new_profile(self, profile: Profile) -> Profile:
        profile.delete_meta()
        profile.delete_profile_id()
        res = self._post_request(
            url=f"{self._base_url}/profiles", data=profile.settings
        )
        return Profile(settings=res)

    def delete_profile(self, profile_id: str):
        res = self._delete_request(f"{self._base_url}/profiles/{profile_id}")
        # print(res)
        return res

    def _list_report_configs(self, params: dict = None) -> List[ReportConfig]:
        res = self._get_request(url=f"{self._base_url}/report-configs", params=params)
        return [ReportConfig(data=rdata) for rdata in res["data"]]

    def list_organisation_report_configs(self) -> List[ReportConfig]:
        return self._list_report_configs()

    def list_group_report_configs(self, group_id: str) -> List[ReportConfig]:
        return self._list_report_configs(params={"groupId": group_id})

    def list_account_report_configs(self, acct_id: str) -> List[ReportConfig]:
        return self._list_report_configs(params={"accountId": acct_id})

    def _create_report_config(
        self,
        report_conf: Dict[str, Any],
        acct_id: str = None,
        group_id: str = None,
    ) -> Dict[str, Any]:
        data: Dict[str, Any] = {"data": {"attributes": {"configuration": report_conf}}}
        if acct_id:
            data["data"]["attributes"]["accountId"] = acct_id
        elif group_id:
            data["data"]["attributes"]["groupId"] = group_id

        res = self._post_request(url=f"{self._base_url}/report-configs", data=data)
        return res["data"]

    def create_organisation_report_config(self, report_conf: Dict[str, Any]):
        return self._create_report_config(report_conf=report_conf)

    def create_group_report_config(self, report_conf: Dict[str, Any], group_id: str):
        return self._create_report_config(report_conf=report_conf, group_id=group_id)

    def create_account_report_config(self, report_conf: Dict[str, Any], acct_id: str):
        return self._create_report_config(report_conf=report_conf, acct_id=acct_id)

    def delete_report_config(self, report_conf_id: str):
        res = self._delete_request(f"{self._base_url}/report-configs/{report_conf_id}")
        return res


class ConformityAPIBaseDecorator:
    def __init__(self, api: ConformityAPI) -> None:
        self._api = api

    @property
    def api(self) -> ConformityAPI:
        return self._api

    def __getattr__(self, name):
        return getattr(self._api, name)


class LegacyConformityAPI(ConformityAPIBaseDecorator):
    def __init__(self, api: ConformityAPI) -> None:
        super().__init__(api)
        self._validate_api()

    def _validate_api(self):
        user = self.current_user()
        if user.is_cloud_one_user:
            raise ConformityError("Not a valid Legacy Conformity API URL")


class CloudOneConformityAPI(ConformityAPIBaseDecorator):
    def __init__(self, api: ConformityAPI) -> None:
        super().__init__(api)
        self._validate_api()

    def _validate_api(self):
        user = self.current_user()
        if not user.is_cloud_one_user:
            raise ConformityError("Not a valid Cloud One Conformity API URL")


class WorkaroundFixConformityAPI(ConformityAPIBaseDecorator):
    def __init__(self, api: ConformityAPI) -> None:
        super().__init__(api)
        self._already_tried_to_access_users = False
        self._successfully_accessed_users = False

    def get_all_users(self) -> List[User]:
        try:
            self._already_tried_to_access_users = True
            users = self.api.get_all_users()
            self._successfully_accessed_users = True
            return users
        except Exception as e:
            self._successfully_accessed_users = True
            raise e

    def _confirmed_not_a_permission_error(self) -> bool:
        # print("Confirming if not a permission error.")
        if not self._already_tried_to_access_users:
            try:
                # print("Trying to access users..")
                self.get_all_users()
                return True
            except Exception:
                return False
        return self._successfully_accessed_users

    def _return_empty_obj_or_raise_error(self, empty_obj, e: Exception):
        if self._confirmed_not_a_permission_error():
            # print("Not really a permission error")
            return empty_obj
        else:
            # print("It is indeed a permission error!")
            raise e

    def list_groups(self, include_group_types: List[str] = None) -> List[Group]:
        try:
            return self.api.list_groups(include_group_types=include_group_types)
        except ConformityForbiddenError as e:
            return self._return_empty_obj_or_raise_error([], e)

    def get_custom_profiles(self) -> List[Profile]:
        try:
            return self.api.get_custom_profiles()
        except ConformityForbiddenError as e:
            return self._return_empty_obj_or_raise_error([], e)

    def list_organisation_report_configs(self) -> List[ReportConfig]:
        try:
            return self.api.list_organisation_report_configs()
        except ConformityForbiddenError as e:
            return self._return_empty_obj_or_raise_error([], e)

    def get_organisation_profile(self, include_rule_settings=False) -> Profile:
        try:
            return self.api.get_organisation_profile(
                include_rule_settings=include_rule_settings
            )
        except ConformityForbiddenError as e:
            org_id = self.get_organisation_id()
            empty_profile = DefaultConformityAPI.create_empty_organisation_profile(
                org_id=org_id, has_empty_included_field=False
            )
            return self._return_empty_obj_or_raise_error(empty_profile, e)

    def list_accounts(self) -> List[Account]:
        accts = self.api.list_accounts()
        if accts is None:
            accts = []
        return accts

    def get_communication_settings(self, acct_id: str) -> List[CommunicationSettings]:
        com_settings = self.api.get_communication_settings(acct_id=acct_id)
        if com_settings is None:
            com_settings = []
        return com_settings

    def update_account_rule_setting(
        self, acct_id: str, rule_id: str, setting: dict, note: str = "Copied from API"
    ):
        try:
            self.api.update_account_rule_setting(
                acct_id=acct_id, rule_id=rule_id, setting=setting, note=note
            )
        except ConformityClientError as e:
            if (
                str(e).startswith("422 Client Error")
                and "`exceptions` is not configurable" in e.details
            ):
                # print(f"Retrying to update rule {rule_id}, now removing exceptions")
                setting.pop("exceptions", None)
                self.api.update_account_rule_setting(
                    acct_id=acct_id, rule_id=rule_id, setting=setting, note=note
                )
            else:
                raise e



import json
from typing import Any, Dict, List, Optional, Union

from deepdiff import DeepDiff, DeepHash


class User:
    ROLE_ADMIN = "ADMIN"
    ROLE_USER = "USER"
    ROLE_READ_ONLY = "READ_ONLY"

    def __init__(
        self,
        user_id: str,
        email: str,
        first_name: str,
        last_name: str,
        role: str,
        mobile_number="",
        is_mobile_verified=False,
        is_cloud_one_user=False,
    ) -> None:
        self.user_id = user_id
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.role = role
        self.mobile_number = mobile_number
        self.is_mobile_verified = is_mobile_verified
        self.is_cloud_one_user = is_cloud_one_user

    def __hash__(self) -> int:
        return hash(self.email)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, User):
            return False
        other: User = other
        return self.email == other.email


class Group:

    GROUP_TYPE_MANAGED_GROUP = "MANAGED_GROUP"
    GROUP_TYPE_USER_DEFINED = ""

    def __init__(
        self,
        group_id: str,
        name: str,
        tags: List[str] = None,
        group_type: str = None,
        cloud_type: str = None,
        cloud_data: dict = None,
    ) -> None:
        self.group_id = group_id
        self.name = name
        self.tags = [] if tags is None else tags
        self._tags = tuple() if tags is None else tuple(sorted(tags))
        self.group_type = group_type
        self.cloud_type = cloud_type
        self.cloud_data = cloud_data

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Group):
            return False
        other: Group = other

        return self.name == other.name and self._tags == other._tags

    def __str__(self) -> str:
        fields = vars(self)
        del fields["_tags"]
        return json.dumps(fields, indent=4)


class CommunicationSettings:
    def __init__(
        self,
        com_setting_id: str,
        channel: str,
        enabled: bool,
        filter: dict,
        configuration: dict,
    ) -> None:
        self.com_setting_id = com_setting_id
        self.channel = channel
        self.enabled = enabled
        self.filter = filter
        self.configuration = configuration
        self._obj = {
            "channel": channel,
            "filter": filter,
            "configuration": configuration,
        }

    def __hash__(self) -> int:
        dh = DeepHash(self._obj)[self._obj]
        return hash(dh)

    def __eq__(self, other: Any) -> bool:
        diff = DeepDiff(self._obj, other._obj, ignore_order=True)
        return len(diff) == 0

    def __str__(self) -> str:
        fields = vars(self)
        del fields["_obj"]
        return json.dumps(fields, indent=4)


class Note:
    def __init__(self, note: str, created_by: str, created_ts: int) -> None:
        self.note = note
        self.created_by = created_by
        self.created_ts = created_ts

    def __str__(self) -> str:
        return json.dumps(vars(self), indent=4)


class Check:
    def __init__(
        self,
        check_id: str,
        acct_id: str,
        rule_id: str,
        service: str,
        region: str,
        resource: str,
        resource_name: str,
        message: str,
        suppressed: Optional[bool],
        suppressed_until: Optional[int],
        notes: List[Note] = None,
    ) -> None:
        self.check_id = check_id
        self.acct_id = acct_id
        self.rule_id = rule_id
        self.service = service
        self.region = region
        self.resource = resource
        self.resource_name = resource_name
        self.message = message
        self.suppressed = suppressed
        self.suppressed_until = suppressed_until
        self.notes = notes if notes is not None else []

    def __hash__(self) -> int:
        return hash(f"{self.rule_id}|{self.resource}")

    def __eq__(self, other: Any) -> bool:
        """
        The equality check is based from how the checkId is composed according
        to the Conformity's API documentation for Custom Check.
        checkID = ccc:accountId:ruleId:service:region:resourceId

        However, for migration purposes, accountId isn't used in this equality
        check because the source and the destination account when migrating
        suppressed check won't have the same accountId. To make sure the
        comparison is accurate, comparison must only be done between two checks
        that come from the same accounts (i.e. 2nd account is the account migrated
        from the 1st account)
        """
        if not isinstance(other, Check):
            return False
        other: Check = other

        return (
            self.rule_id == other.rule_id
            and self.service == other.service
            and self.region == other.region
            and self.resource == other.resource
        )

    def __str__(self) -> str:
        return json.dumps(vars(self), indent=4)


class Rule:
    def __init__(self, setting: dict, notes: List[Note] = None) -> None:
        self.setting = setting
        self.notes = notes if notes is not None else []
        self.rule_id = setting["id"]
        self.enabled = setting["enabled"]
        self.configured = setting.get("configured", False)

    def __hash__(self) -> int:
        return hash(self.rule_id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Rule):
            return False
        other: Rule = other
        return self.rule_id == other.rule_id


class Profile:
    def __init__(self, settings: dict) -> None:
        self.settings = settings
        # data = settings["data"]
        # attrib = data["attributes"]
        # self.profile_id = data.get("id", "")
        # self.name = attrib["name"]
        # self.description = attrib.get("description", "")

    @property
    def profile_id(self) -> str:
        return self.settings["data"].get("id", "")

    @profile_id.setter
    def profile_id(self, profile_id: str) -> None:
        self.settings["data"]["id"] = profile_id

    @property
    def name(self) -> str:
        return self.settings["data"]["attributes"]["name"]

    @name.setter
    def name(self, name: str) -> None:
        self.settings["data"]["attributes"]["name"] = name

    @property
    def description(self) -> str:
        return self.settings["data"]["attributes"].get("description", "")

    @description.setter
    def description(self, description: str) -> None:
        self.settings["data"]["attributes"]["description"] = description

    @property
    def included_rules(self) -> Union[List[dict], None]:
        return self.settings.get("included")

    def delete_profile_id(self) -> None:
        if "id" in self.settings["data"]:
            del self.settings["data"]["id"]

    def delete_meta(self) -> None:
        if "meta" in self.settings:
            del self.settings["meta"]

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Profile):
            return False
        other: Profile = other
        return self.name == other.name


class Account:
    def __init__(self, acct_data: Dict[str, Any]) -> None:
        self.data = acct_data

    @property
    def attributes(self) -> Dict[str, Any]:
        return self.data["attributes"]

    @property
    def account_id(self) -> str:
        return self.data["id"]

    @property
    def name(self) -> str:
        return self.attributes["name"]

    @property
    def environment(self) -> str:
        env = self.attributes.get("environment", "")
        if env is None:
            env = ""
        return env

    @property
    def cloud_type(self) -> str:
        return self.attributes["cloud-type"]

    @property
    def tags(self) -> List[str]:
        return self.attributes.get("tags", [])

    @property
    def managed_group_id(self) -> str:
        return self.attributes["managed-group-id"]

    @property
    def security_package(self) -> bool:
        return self.attributes["security-package"]

    @property
    def organisation_id(self) -> str:
        return self.data["relationships"]["organisation"]["data"]["id"]


class AccountDetails(Account):
    def __init__(self, acct_data: Dict[str, Any]) -> None:
        super().__init__(acct_data=acct_data)

    @property
    def rules(self) -> List[Rule]:
        rsettings = self.attributes["settings"].get("rules")
        if not rsettings:
            return []
        return [Rule(setting=rule_setting) for rule_setting in rsettings]

    @property
    def bot_settings(self) -> Union[Dict[str, Any], None]:
        return self.attributes["settings"].get("bot")

    @property
    def bot_status(self) -> Union[str, None]:
        return self.attributes.get("bot-status")


class ReportConfig:
    def __init__(self, data: Dict[str, Any]) -> None:
        self.data = data

    @property
    def report_config_id(self) -> str:
        return self.data["id"]

    @property
    def enabled(self) -> bool:
        return self.data["attributes"]["enabled"]

    @property
    def configuration(self) -> Dict[str, Any]:
        return self.data["attributes"]["configuration"]

    @property
    def title(self) -> str:
        return self.configuration["title"]

    @property
    def description(self) -> str:
        return self.configuration["description"]

    @property
    def scheduled(self) -> bool:
        return self.configuration["scheduled"]

    @property
    def is_account_level(self) -> bool:
        return self.data["attributes"]["is-account-level"]

    @property
    def is_group_level(self) -> bool:
        return self.data["attributes"]["is-group-level"]

    @property
    def is_organisation_level(self) -> bool:
        return self.data["attributes"]["is-organisation-level"]

    def __hash__(self) -> int:
        return hash(self.title)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ReportConfig):
            return False
        other: ReportConfig = other
        return self.title == other.title



import csv
import json
import multiprocessing as mp
import time
from dataclasses import dataclass
from typing import Iterable, Tuple

import boto3
import click
from mypy_boto3_cloudformation import CloudFormationClient
from mypy_boto3_cloudformation.type_defs import (
    DescribeStacksOutputTypeDef,
    ParameterTypeDef,
    UpdateStackOutputTypeDef,
)

from .di import c1_conformity_api, legacy_conformity_api


@dataclass
class LegacyConformityAWSAccountInfo:
    account_name: str
    aws_account_number: str
    old_external_id: str


def get_legacy_conformity_aws_accounts_info() -> Iterable[
    LegacyConformityAWSAccountInfo
]:
    legacy_api = legacy_conformity_api()
    old_external_id = legacy_api.get_organisation_external_id()
    accts = [acct for acct in legacy_api.list_accounts() if acct.cloud_type == "aws"]
    for acct in accts:
        yield LegacyConformityAWSAccountInfo(
            account_name=acct.name,
            aws_account_number=acct.attributes["awsaccount-id"],
            old_external_id=old_external_id,
        )


@click.group()
@click.pass_context
def cli(ctx):
    ctx.ensure_object(dict)
    # ctx.obj["region"] = region
    # ctx.obj["profile"] = profile


@cli.command(
    "generate-csv",
    help="Creates a csv file containing AWS accounts to be used for 'update-stack --csv-file' command option.",
)
@click.argument("csv-file")
def generate_csv(csv_file: str):
    print(f"Generating CSV: {csv_file}")
    with open(csv_file, newline="", mode="w") as fh:
        csvw = csv.DictWriter(
            fh,
            fieldnames=[
                "Account Name",
                "AWS Account Number",
                "Old ExternalID",
                "Conformity Stack Name",
                "Conformity Stack Region",
                "AWS_PROFILE",
                "AWS_ACCESS_KEY_ID",
                "AWS_SECRET_ACCESS_KEY",
                "AWS_SESSION_TOKEN",
                "Cross-Account Role Name",
            ],
            dialect="excel",
        )
        csvw.writeheader()

        accts = get_legacy_conformity_aws_accounts_info()
        for acct in accts:
            csvw.writerow(
                {
                    "Account Name": acct.account_name,
                    "AWS Account Number": acct.aws_account_number,
                    "Old ExternalID": acct.old_external_id,
                    "Conformity Stack Name": "CloudConformity",
                    "Conformity Stack Region": "us-east-1",
                    "AWS_PROFILE": "",
                    "AWS_ACCESS_KEY_ID": "",
                    "AWS_SECRET_ACCESS_KEY": "",
                    "AWS_SESSION_TOKEN": "",
                    "Cross-Account Role Name": "",
                }
            )
    print("Done!")
    print(
        """
    You may now open and edit the csv especially the credentials column(s).
    You can either edit AWS_PROFILE or AWS_ACCESS_KEY_ID/AWS_SECRET_ACCESS_KEY/AWS_SESSION_TOKEN
    If you specify both, then it will ignore AWS_PROFILE.
    AWS_SESSION_TOKEN is optional - you will need this when the credentials are temporary ones.

    The values in the columns 'Conformity Stack Name' and 'Conformity Stack Region' are the default ones.
    You may edit them as necessary.
"""
    )


@cli.command("update-stack", help="Updates ExternalID of Cloud Conformity Stack")
@click.option(
    "--stack-name",
    default="CloudConformity",
    show_default=True,
    required=False,
    help="Name of the Cloud Conformity Stack",
)
@click.option(
    "--external-id",
    type=str,
    required=False,
    help="New value for the Stack paramater ExternalID. If not specified, it will use the ExternalId from Cloud One Conformity.",
)
@click.option(
    "--csv-file",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
    required=False,
    help="An optional CSV file containing AWS accounts' credentials and stack information. Non-empty field values in the CSV file will override whatever values entered in the CLI options.",
)
@click.option(
    "--region",
    type=str,
    envvar="AWS_DEFAULT_REGION",
    show_envvar=True,
    required=False,
    show_default=True,
    default="us-east-1",
    help="Region where Cloud Conformity Stack is deployed",
)
@click.option(
    "--profile",
    type=str,
    envvar="AWS_PROFILE",
    show_envvar=True,
    required=False,
    default=None,
    help="AWS credentials/config profile to use",
)
@click.option(
    "--access-key",
    type=str,
    envvar="AWS_ACCESS_KEY_ID",
    show_envvar=True,
    required=False,
    default=None,
    help="AWS Access Key",
)
@click.option(
    "--secret-key",
    type=str,
    envvar="AWS_SECRET_ACCESS_KEY",
    show_envvar=True,
    required=False,
    default=None,
    help="AWS Secret Key",
)
@click.option(
    "--session-token",
    type=str,
    envvar="AWS_SESSION_TOKEN",
    show_envvar=True,
    required=False,
    default=None,
    help="AWS Session Token",
)
@click.option(
    "--cross-account-role-name",
    type=str,
    required=False,
    default=None,
    help="Cross-Account Role name (e.g. OrganizationAccountAccessRole). The role should at least have the permissions necessary to update the Conformity stack.",
)
@click.pass_context
def update_stack(
    ctx,
    stack_name: str,
    external_id: str,
    csv_file: str,
    region: str,
    profile: str,
    access_key: str,
    secret_key: str,
    session_token: str,
    cross_account_role_name: str,
):
    # region = ctx.obj["region"]
    # profile = ctx.obj["profile"]
    if not external_id:
        print("Retrieving ExternalId from Cloud One Conformity")
        external_id = c1_conformity_api().get_organisation_external_id()
        print(f"ExternalId: {external_id}")

    accts: Iterable[AccountStackInfo]
    if csv_file:
        accts = read_csv_file(csv_file=csv_file)
        accts = _fill_acct_with_defaults(
            accts=accts,
            default_stack_name=stack_name,
            default_profile=profile,
            default_region=region,
            default_access_key=access_key,
            default_secret_key=secret_key,
            default_session_token=session_token,
            default_cross_account_role_name=cross_account_role_name,
        )
    else:
        accts = (
            AccountStackInfo(
                account_name=acct.account_name,
                aws_account_number=acct.aws_account_number,
                old_external_id=acct.old_external_id,
                stack_name=stack_name,
                stack_region=region,
                aws_profile=profile,
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                aws_session_token=session_token,
                cross_account_role_name=cross_account_role_name,
            )
            for acct in get_legacy_conformity_aws_accounts_info()
        )

    accts = list(accts)
    proc_count = min(len(accts), 10)
    with mp.Pool(processes=proc_count) as pool:
        params = _update_stack_params(accts=accts, external_id=external_id)
        pool.map(_update_stack_worker, params)


@dataclass
class AccountStackInfo:
    account_name: str
    aws_account_number: str
    old_external_id: str
    stack_name: str
    stack_region: str
    aws_profile: str
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str
    cross_account_role_name: str


def _update_stack_params(
    accts: Iterable[AccountStackInfo],
    external_id: str,
) -> Iterable[Tuple[AccountStackInfo, str]]:
    for acct in accts:
        yield (acct, external_id)


def _update_stack_worker(params: Tuple[AccountStackInfo, str]):
    # print(f"[Process: {os.getpid()}] Params: {params}")
    acct, external_id = params
    try:
        _update_stack(acct=acct, external_id=external_id)
    except Exception as e:
        print(f"Failed to update stack for {acct.account_name}. Error: {e}")


def _fill_acct_with_defaults(
    accts: Iterable[AccountStackInfo],
    default_stack_name: str,
    default_profile: str,
    default_region: str,
    default_access_key: str,
    default_secret_key: str,
    default_session_token: str,
    default_cross_account_role_name: str,
) -> Iterable[AccountStackInfo]:
    for acct in accts:
        stack_name = acct.stack_name if acct.stack_name else default_stack_name
        stack_region = acct.stack_region if acct.stack_region else default_region
        aws_profile = acct.aws_profile if acct.aws_profile else default_profile
        access_key = (
            acct.aws_access_key_id if acct.aws_access_key_id else default_access_key
        )
        secret_key = (
            acct.aws_secret_access_key
            if acct.aws_secret_access_key
            else default_secret_key
        )
        session_token = (
            acct.aws_session_token if acct.aws_session_token else default_session_token
        )
        cross_account_role_name = (
            acct.cross_account_role_name
            if acct.cross_account_role_name
            else default_cross_account_role_name
        )
        yield AccountStackInfo(
            account_name=acct.account_name,
            aws_account_number=acct.aws_account_number,
            old_external_id=acct.old_external_id,
            stack_name=stack_name,
            stack_region=stack_region,
            aws_profile=aws_profile,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=session_token,
            cross_account_role_name=cross_account_role_name,
        )


def read_csv_file(csv_file: str) -> Iterable[AccountStackInfo]:
    with open(csv_file, newline="", mode="r") as fh:
        csvr = csv.DictReader(fh, dialect="excel")
        for rec in csvr:
            acct = AccountStackInfo(
                account_name=rec["Account Name"],
                aws_account_number=rec["AWS Account Number"].strip(),
                old_external_id=rec["Old ExternalID"].strip(),
                stack_name=rec["Conformity Stack Name"].strip(),
                stack_region=rec["Conformity Stack Region"].strip(),
                aws_profile=rec["AWS_PROFILE"].strip(),
                aws_access_key_id=rec["AWS_ACCESS_KEY_ID"].strip(),
                aws_secret_access_key=rec["AWS_SECRET_ACCESS_KEY"].strip(),
                aws_session_token=rec["AWS_SESSION_TOKEN"].strip(),
                cross_account_role_name=rec.get("Cross-Account Role Name", "").strip(),
            )
            yield acct


def _get_sess_acct_number(sess: boto3.Session) -> str:
    sts = sess.client("sts")
    return sts.get_caller_identity()["Account"]


def _get_cross_acct_sess(sess: boto3.Session, aws_acct_num: str, role_name: str):
    sess_acct_num = _get_sess_acct_number(sess)
    if sess_acct_num == aws_acct_num:
        return sess

    sts = sess.client("sts")
    resp = sts.assume_role(
        RoleArn=f"arn:aws:iam::{aws_acct_num}:role/{role_name}",
        RoleSessionName=f"cross_acct_sess_{sess_acct_num}",
        DurationSeconds=3600,
    )

    creds = resp["Credentials"]

    return boto3.Session(
        region_name=sess.region_name,
        aws_access_key_id=creds["AccessKeyId"],
        aws_secret_access_key=creds["SecretAccessKey"],
        aws_session_token=creds["SessionToken"],
    )


def _update_stack(acct: AccountStackInfo, external_id: str):

    if acct.aws_access_key_id and acct.aws_secret_access_key:
        session_token = acct.aws_session_token if acct.aws_session_token else None
        sess = boto3.Session(
            region_name=acct.stack_region,
            aws_access_key_id=acct.aws_access_key_id,
            aws_secret_access_key=acct.aws_secret_access_key,
            aws_session_token=session_token,
        )
    else:
        sess = boto3.Session(
            profile_name=acct.aws_profile, region_name=acct.stack_region
        )

    if acct.cross_account_role_name:
        sess = _get_cross_acct_sess(
            sess=sess,
            aws_acct_num=acct.aws_account_number,
            role_name=acct.cross_account_role_name,
        )

    sess_acct_num = _get_sess_acct_number(sess)
    if sess_acct_num != acct.aws_account_number:
        print(
            f"AWS credentials not for this account AWS={acct.aws_account_number} ({acct.account_name})"
        )
        return

    cfn = sess.client("cloudformation", region_name=acct.stack_region)
    old_external_id = get_stack_external_id(cfn=cfn, stack_name=acct.stack_name)

    acct_info = (
        f" [AWS={acct.aws_account_number} ({acct.account_name})]"
        if acct.aws_account_number
        else ""
    )

    if external_id == old_external_id:
        print(
            f"[Update stack skipped]{acct_info} [{old_external_id} --> {external_id}]",
            flush=True,
        )
        return

    print(
        f"[Update stack started]{acct_info} [{old_external_id} --> {external_id}]",
        flush=True,
    )

    params = [
        ParameterTypeDef(
            ParameterKey="AccountId",
            UsePreviousValue=True,
        ),
        ParameterTypeDef(
            ParameterKey="ExternalId",
            ParameterValue=external_id,
        ),
    ]
    res: UpdateStackOutputTypeDef = cfn.update_stack(
        StackName=acct.stack_name,
        UsePreviousTemplate=True,
        Parameters=params,
        Capabilities=["CAPABILITY_NAMED_IAM"],
    )
    stack_id = res["StackId"]

    (is_successful, reason) = wait_for_update_stack(
        cfn=cfn, stack_id=stack_id, check_interval_in_secs=5
    )
    if is_successful:
        print(
            f"[Update stack success]{acct_info} [{old_external_id} --> {external_id}]"
        )
    else:
        print(
            f"[Update stack failed! Reason={reason}]{acct_info} [{old_external_id} --> {external_id}]"
        )


def get_stack_external_id(cfn: CloudFormationClient, stack_name: str) -> str:
    res: DescribeStacksOutputTypeDef = cfn.describe_stacks(StackName=stack_name)
    stack = res["Stacks"][0]
    params = stack["Parameters"]
    for param in params:
        if param["ParameterKey"] == "ExternalId":
            return param["ParameterValue"]
    return ""


def wait_for_update_stack(
    cfn: CloudFormationClient, stack_id: str, check_interval_in_secs=5
) -> Tuple[bool, str]:
    is_success = False
    reason = ""
    while True:
        res: DescribeStacksOutputTypeDef = cfn.describe_stacks(StackName=stack_id)
        stack = res["Stacks"][0]
        status = stack["StackStatus"]
        if status == "UPDATE_IN_PROGRESS":
            time.sleep(check_interval_in_secs)
            continue
        reason = stack.get("StackStatusReason", "")
        reason = f"({status}) {reason}"
        if status in {"UPDATE_COMPLETE", "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS"}:
            is_success = True
            break
        else:
            pretty_print(res)
            is_success = False
            break
    return is_success, reason


def pretty_print(obj):
    print(json.dumps(obj, indent=4, default=str))


if __name__ == "__main__":
    cli()



import os
from abc import ABCMeta, abstractmethod
from typing import List, Tuple, Union

from PyInquirer import prompt

from conformity_migration_tool.utils import str2bool

from .conformity_api import ConformityAPI
from .models import Account


class CloudAccountAdder(metaclass=ABCMeta):
    @abstractmethod
    def account_exists(
        self, c1_accts: List[Account], acct: Account
    ) -> Tuple[bool, str]:
        pass

    @abstractmethod
    def account_add(self, acct: Account) -> str:
        pass


class AWSCloudAccountAdder(CloudAccountAdder):
    def __init__(self, legacy_api: ConformityAPI, c1_api: ConformityAPI) -> None:
        self.legacy_api = legacy_api
        self.c1_api = c1_api

    def _account_uniq_attrib(self, acct: Account) -> str:
        aws_acct_num = acct.attributes.get("awsaccount-id", "")
        if not aws_acct_num:
            print(f"No AWS account number for: {acct.name} {acct.environment}")
        return aws_acct_num

    def account_exists(
        self, c1_accts: List[Account], acct: Account
    ) -> Tuple[bool, str]:
        for c1_acct in c1_accts:
            if c1_acct.cloud_type != "aws":
                continue
            if self._account_uniq_attrib(acct) == self._account_uniq_attrib(c1_acct):
                return True, c1_acct.account_id
        return False, ""

    def account_add(self, acct: Account) -> str:
        name = acct.name
        environment = acct.environment
        legacy_acct_id = acct.account_id

        c1_external_id = self.c1_api.get_organisation_external_id()

        aws_acct_num = acct.attributes["awsaccount-id"]

        access_conf = self.legacy_api.get_account_access_configuration(
            acct_id=legacy_acct_id
        )
        role_arn = access_conf["roleArn"]
        old_external_id = access_conf["externalId"]

        skip_aws_prompt = str2bool(os.getenv("SKIP_AWS_PROMPT", "False"))
        if not skip_aws_prompt:
            self.show_update_stack_external_id_instructions(
                aws_acct_num=aws_acct_num,
                old_external_id=old_external_id,
                new_external_id=c1_external_id,
            )
            prompt_continue()

        res = self.c1_api.add_aws_account(
            name=name,
            environment=environment,
            role_arn=role_arn,
            external_id=c1_external_id,
        )

        return res["id"]

    @staticmethod
    def show_update_stack_external_id_instructions(
        aws_acct_num, old_external_id, new_external_id
    ):
        print(
            f"""
Please do the following steps to grant CloudOne Conformity access to your AWS account:
    1. Sign in to AWS console for AWS Account {aws_acct_num}
    2. Go to CloudFormation and find stack name CloudConformity
    3. Click Update button to edit stack.
    4. Under Prepare Template, choose: Use current template, and click Next
    5. Under Parameters, change value of ExternalID to the new one below:
        Old Value: {old_external_id}
        New Value: {new_external_id}
"""
        )


class AzureCloudAccountAdder(CloudAccountAdder):
    def __init__(self, legacy_api: ConformityAPI, c1_api: ConformityAPI) -> None:
        self.legacy_api = legacy_api
        self.c1_api = c1_api

    def _account_uniq_attrib(self, acct: Account) -> str:
        return acct.attributes["cloud-data"]["azure"]["subscriptionId"]

    def account_exists(
        self, c1_accts: List[Account], acct: Account
    ) -> Tuple[bool, str]:
        for c1_acct in c1_accts:
            if c1_acct.cloud_type != "azure":
                continue
            if self._account_uniq_attrib(acct) == self._account_uniq_attrib(c1_acct):
                return True, c1_acct.account_id
        return False, ""

    def account_add(self, acct: Account) -> str:
        name = acct.name
        environment = acct.environment

        azure_sub_id = acct.attributes["cloud-data"]["azure"]["subscriptionId"]
        group_id = acct.managed_group_id

        res = self.legacy_api.get_group_details(group_id=group_id)
        gattrib = res["attributes"]
        azure_data = gattrib["cloud-data"]["azure"]
        active_directory_id = azure_data["directoryId"]

        res = self.c1_api.add_azure_subscription(
            name=name,
            environment=environment,
            subscription_id=azure_sub_id,
            active_directory_id=active_directory_id,
        )
        # print(res)
        return res["id"]


def prompt_continue():
    questions = [
        {
            "type": "confirm",
            "message": "Do you want to continue?",
            "name": "continue",
            "default": False,
        },
    ]
    while (prompt(questions=questions))["continue"] is False:
        pass


def get_cloud_account_adder(
    cloud_type: str, legacy_api: ConformityAPI, c1_api: ConformityAPI
) -> Union[CloudAccountAdder, None]:
    if cloud_type == "aws":
        return AWSCloudAccountAdder(legacy_api, c1_api)
    if cloud_type == "azure":
        return AzureCloudAccountAdder(legacy_api, c1_api)
    return None



from setuptools import setup

REQUIRES = [
    "typing_extensions>=4.0.0",
    "requests>=2.0.0,<3.0.0",
    "PyYAML>=6.0,<7.0",
    "deepdiff>=5.7.0,<6.0.0",
    "backoff>=1.11.1",
    "vcrpy>=4.0.0,<5.0.0",
    # cli dependencies below
    "PyInquirer>=1.0.3,<2.0.0",
    "click>=8.0.0,<9.0.0",
    "importlib_metadata; python_version < '3.8'",
    "boto3>=1.0.0,<2.0.0",
    "boto3-stubs[cloudformation]",
]
setup(name="conformity-migration-tool", install_requires=REQUIRES)



import logging


class Logger:
    def info(self, msg: object, *args, file_only=False, **kwargs) -> None:
        ...

    def warn(self, msg: object, *args, file_only=False, **kwargs) -> None:
        ...

    def debug(self, msg: object, *args, file_only=False, **kwargs) -> None:
        ...

    def error(self, msg: object, *args, file_only=False, **kwargs) -> None:
        ...

    def exception(self, msg: object, *args, file_only=False, **kwargs) -> None:
        ...


class AppLogger(Logger):
    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger

    def _prepare_for_log(self, kwargs: dict, file_only=False):
        # remove these params from converted print statements
        kwargs.pop("end", None)
        kwargs.pop("flush", None)

        extra: dict = kwargs.setdefault("extra", dict())
        extra["file_only"] = file_only

    def info(self, msg: object, *args, file_only=False, **kwargs) -> None:
        self._prepare_for_log(kwargs, file_only=file_only)
        return self.logger.info(msg, *args, **kwargs)

    def warn(self, msg: object, *args, file_only=False, **kwargs) -> None:
        self._prepare_for_log(kwargs, file_only=file_only)
        return self.logger.warning(msg, *args, **kwargs)

    def debug(self, msg: object, *args, file_only=False, **kwargs) -> None:
        self._prepare_for_log(kwargs, file_only=file_only)
        return self.logger.debug(msg, *args, **kwargs)

    def error(self, msg: object, *args, file_only=False, **kwargs) -> None:
        self._prepare_for_log(kwargs, file_only=file_only)
        return self.logger.error(msg, *args, **kwargs)

    def exception(self, msg: object, *args, file_only=False, **kwargs) -> None:
        self._prepare_for_log(kwargs, file_only=file_only)
        return self.logger.exception(msg, *args, **kwargs)


class NoStrackTraceExceptionFormatter(logging.Formatter):
    def formatException(self, exc_info) -> str:
        return str(exc_info[1])

    def format(self, record: logging.LogRecord):
        # clears cached exc_text formatted by other Formatter.formatException(record.exc_info)
        record.exc_text = ""
        return super().format(record=record)


class WithStrackTraceExceptionFormatter(logging.Formatter):
    def formatException(self, exc_info) -> str:
        return super().formatException(exc_info)

    def format(self, record: logging.LogRecord):
        # clears cached exc_text formatted by other Formatter.formatException(record.exc_info)
        record.exc_text = ""
        return super().format(record=record)
